"""Value masks for Noisemaker. Used when generating value noise or glyph maps."""

import random

import numpy as np

from noisemaker.constants import ValueMask
from noisemaker.simplex import simplex


#: Hard-coded masks
Masks = {
    ValueMask.chess: [[0, 1], [1, 0]],

    ValueMask.waffle: [[0, 1], [1, 1]],

    ValueMask.h_bar: [
        [1, 1],
        [0, 0],
    ],

    ValueMask.v_bar: [
        [1, 0],
        [1, 0],
    ],

    ValueMask.h_hex: [
        [0, 1, 0, 0],
        [0, 0, 0, 1],
        [0, 0, 0, 0],
        [0, 0, 0, 1],
        [0, 1, 0, 0],
        [0, 0, 0, 0],
    ],

    ValueMask.v_hex: [
        [0, 1, 0, 1, 0, 0],
        [0, 0, 0, 0, 0, 0],
        [1, 0, 0, 0, 1, 0],
        [0, 0, 0, 0, 0, 0]
    ],

    ValueMask.h_tri: [
        [0, 1],
        [0, 0],
        [1, 0],
        [0, 0]
    ],

    ValueMask.v_tri: [
        [1, 0, 0, 0],
        [0, 0, 1, 0]
    ],

    ValueMask.square: [
        [0, 0, 0, 0],
        [0, 1, 1, 0],
        [0, 1, 1, 0],
        [0, 0, 0, 0]
    ],

    ValueMask.grid: [
        [[1, 1, 1]] * 5 + [[0, 0, 0]],
        [[1, 1, 1]] * 5 + [[0, 0, 0]],
        [[1, 1, 1]] * 5 + [[0, 0, 0]],
        [[1, 1, 1]] * 5 + [[0, 0, 0]],
        [[1, 1, 1]] * 5 + [[0, 0, 0]],
        [[0, 0, 0]] * 6,
    ],

    ValueMask.zero: [
        [ 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 1, 1, 0 ],
        [ 1, 0, 1, 0, 1, 0 ],
        [ 1, 1, 0, 0, 1, 0 ],
        [ 0, 1, 1, 1, 0, 0 ]
    ],

    ValueMask.one: [
        [ 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0 ],
        [ 0, 1, 1, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0 ],
        [ 0, 1, 1, 1, 0, 0 ]
    ],

    ValueMask.two: [
        [ 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0 ],
        [ 0, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0 ],
        [ 1, 1, 1, 1, 1, 0 ]
    ],

    ValueMask.three: [
        [ 0, 0, 0, 0, 0, 0 ],
        [ 1, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0 ],
        [ 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0 ],
        [ 1, 1, 1, 1, 0, 0 ]
    ],

    ValueMask.four: [
        [ 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 1, 0, 0 ],
        [ 1, 0, 0, 1, 0, 0 ],
        [ 1, 1, 1, 1, 1, 0 ],
        [ 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0 ]
    ],

    ValueMask.five: [
        [ 0, 0, 0, 0, 0, 0 ],
        [ 1, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0 ],
        [ 1, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0 ],
        [ 1, 1, 1, 1, 0, 0 ]
    ],

    ValueMask.six: [
        [ 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0 ],
        [ 1, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 1, 0 ],
        [ 0, 1, 1, 1, 0, 0 ]
    ],

    ValueMask.seven: [
        [ 0, 0, 0, 0, 0, 0 ],
        [ 1, 1, 1, 1, 1, 0 ],
        [ 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0 ]
    ],

    ValueMask.eight: [
        [ 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 1, 0 ],
        [ 0, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 1, 0 ],
        [ 0, 1, 1, 1, 0, 0 ]
    ],

    ValueMask.nine: [
        [ 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 1, 0 ],
        [ 0, 1, 1, 1, 1, 0 ],
        [ 0, 0, 0, 0, 1, 0 ],
        [ 0, 1, 1, 1, 0, 0 ]
    ],

    ValueMask.a: [
        [ 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 1, 0 ],
        [ 1, 1, 1, 1, 1, 0 ],
        [ 1, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 1, 0 ]
    ],

    ValueMask.b: [
        [ 0, 0, 0, 0, 0, 0 ],
        [ 1, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 1, 0 ],
        [ 1, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 1, 0 ],
        [ 1, 1, 1, 1, 0, 0 ]
    ],

    ValueMask.c: [
        [ 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 1, 1, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0 ],
        [ 0, 1, 1, 1, 1, 0 ]
    ],

    ValueMask.d: [
        [ 0, 0, 0, 0, 0, 0 ],
        [ 1, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 1, 0 ],
        [ 1, 1, 1, 1, 0, 0 ]
    ],

    ValueMask.e: [
        [ 0, 0, 0, 0, 0, 0 ],
        [ 1, 1, 1, 1, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0 ],
        [ 1, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0 ],
        [ 1, 1, 1, 1, 1, 0 ]
    ],

    ValueMask.f: [
        [ 0, 0, 0, 0, 0, 0 ],
        [ 1, 1, 1, 1, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0 ],
        [ 1, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0 ]
    ],

    ValueMask.tromino_i: [
        [ 1, 0, 0, 0 ],
        [ 1, 0, 0, 0 ],
        [ 1, 0, 0, 0 ],
        [ 1, 0, 0, 0 ]
    ],

    ValueMask.tromino_l: [
        [ 0, 0, 0, 0 ],
        [ 0, 1, 0, 0 ],
        [ 0, 1, 0, 0 ],
        [ 0, 1, 1, 0 ]
    ],

    ValueMask.tromino_o: [
        [ 0, 0, 0, 0 ],
        [ 0, 1, 1, 0 ],
        [ 0, 1, 1, 0 ],
        [ 0, 0, 0, 0 ]
    ],

    ValueMask.tromino_s: [
        [ 0, 0, 0, 0 ],
        [ 0, 1, 1, 0 ],
        [ 1, 1, 0, 0 ],
        [ 0, 0, 0, 0 ]
    ],

    ValueMask.halftone_0: [
        [ 0, 0, 0, 0 ],
        [ 0, 0, 0, 0 ],
        [ 0, 0, 0, 0 ],
        [ 0, 0, 0, 0 ]
    ],

    ValueMask.halftone_1: [
        [ 0, 0, 0, 0 ],
        [ 0, 1, 0, 0 ],
        [ 0, 0, 0, 0 ],
        [ 0, 0, 0, 0 ]
    ],

    ValueMask.halftone_2: [
        [ 0, 0, 0, 0 ],
        [ 0, 1, 0, 0 ],
        [ 0, 0, 0, 0 ],
        [ 0, 0, 0, 1 ]
    ],

    ValueMask.halftone_3: [
        [ 0, 0, 0, 0 ],
        [ 0, 1, 0, 1 ],
        [ 0, 0, 0, 0 ],
        [ 0, 1, 0, 1 ],
    ],

    ValueMask.halftone_4: [
        [ 1, 0, 1, 0 ],
        [ 0, 1, 0, 1 ],
        [ 1, 0, 1, 0 ],
        [ 0, 1, 0, 1 ],
    ],

    ValueMask.halftone_5: [
        [ 0, 1, 0, 1 ],
        [ 1, 0, 1, 0 ],
        [ 0, 1, 0, 1 ],
        [ 1, 0, 1, 0 ]
    ],

    ValueMask.halftone_6: [
        [ 1, 1, 1, 1 ],
        [ 1, 0, 1, 0 ],
        [ 1, 1, 1, 1 ],
        [ 1, 0, 1, 0 ]
    ],

    ValueMask.halftone_7: [
        [ 1, 1, 1, 1 ],
        [ 1, 0, 1, 1 ],
        [ 1, 1, 1, 1 ],
        [ 1, 1, 1, 0 ]
    ],

    ValueMask.halftone_8: [
        [ 1, 1, 1, 1 ],
        [ 1, 0, 1, 1 ],
        [ 1, 1, 1, 1 ],
        [ 1, 1, 1, 1 ]
    ],

    ValueMask.halftone_9: [
        [ 1, 1, 1, 1 ],
        [ 1, 1, 1, 1 ],
        [ 1, 1, 1, 1 ],
        [ 1, 1, 1, 1 ]
    ],

    ValueMask.lcd_0: [
        [ 0, 1, 1, 0, 0 ],
        [ 1, 0, 0, 1, 0 ],
        [ 1, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 1, 0 ],
        [ 1, 0, 0, 1, 0 ],
        [ 0, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0 ],
    ],

    ValueMask.lcd_1: [
        [ 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0 ],
    ],

    ValueMask.lcd_2: [
        [ 0, 1, 1, 0, 0 ],
        [ 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 1, 0 ],
        [ 0, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0 ],
        [ 0, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0 ],
    ],

    ValueMask.lcd_3: [
        [ 0, 1, 1, 0, 0 ],
        [ 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 1, 0 ],
        [ 0, 1, 1, 0, 0 ],
        [ 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 1, 0 ],
        [ 0, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0 ],
    ],

    ValueMask.lcd_4: [
        [ 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 1, 0 ],
        [ 1, 0, 0, 1, 0 ],
        [ 0, 1, 1, 0, 0 ],
        [ 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0 ],
    ],

    ValueMask.lcd_5: [
        [ 0, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0 ],
        [ 0, 1, 1, 0, 0 ],
        [ 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 1, 0 ],
        [ 0, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0 ],
    ],

    ValueMask.lcd_6: [
        [ 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0 ],
        [ 0, 1, 1, 0, 0 ],
        [ 1, 0, 0, 1, 0 ],
        [ 1, 0, 0, 1, 0 ],
        [ 0, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0 ],
    ],

    ValueMask.lcd_7: [
        [ 0, 1, 1, 0, 0 ],
        [ 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0 ],
    ],

    ValueMask.lcd_8: [
        [ 0, 1, 1, 0, 0 ],
        [ 1, 0, 0, 1, 0 ],
        [ 1, 0, 0, 1, 0 ],
        [ 0, 1, 1, 0, 0 ],
        [ 1, 0, 0, 1, 0 ],
        [ 1, 0, 0, 1, 0 ],
        [ 0, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0 ],
    ],

    ValueMask.lcd_9: [
        [ 0, 1, 1, 0, 0 ],
        [ 1, 0, 0, 1, 0 ],
        [ 1, 0, 0, 1, 0 ],
        [ 0, 1, 1, 0, 0 ],
        [ 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_0: [
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 1, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 1, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 1, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 1, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 1, 0, 0, 0, 0, 0, 1, 0 ],
        [ 1, 1, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_1: [
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 1, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_2: [
        [ 0, 1, 1, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_3: [
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_4: [
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_5: [
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 1, 0, 0, 0, 0 ],
        [ 0, 1, 1, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_6: [
        [ 0, 0, 0, 0, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_7: [
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_8: [
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_9: [
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 1, 0, 0, 0, 0 ],
        [ 0, 1, 1, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_a: [
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 1, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_b: [
        [ 0, 1, 1, 1, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 1, 0, 0, 0, 0 ],
        [ 0, 1, 1, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_c: [
        [ 0, 0, 0, 0, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_d: [
        [ 0, 1, 1, 1, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 1, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_e: [
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 1, 1, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_f: [
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 1, 1, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_g: [
        [ 0, 0, 0, 0, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 1, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 1, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 1, 0, 0, 1, 0 ],
        [ 0, 1, 1, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_h: [
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_i: [
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_j: [
        [ 0, 0, 0, 0, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_k: [
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 1, 0, 0, 0, 0 ],
        [ 0, 1, 1, 1, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 1, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_l: [
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_m: [
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 1, 0, 0, 0, 0, 0, 1, 1, 0 ],
        [ 1, 0, 1, 0, 0, 0, 1, 0, 1, 0 ],
        [ 1, 0, 0, 1, 0, 1, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 1, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 1, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 1, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_n: [
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 1, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 1, 0, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 1, 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 1, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 1, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 1, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_o: [
        [ 0, 1, 1, 1, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 1, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_p: [
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_q: [
        [ 0, 0, 0, 0, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 1, 0, 1, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 1, 0, 1, 0, 0 ],
        [ 0, 1, 1, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_r: [
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 0, 1, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_s: [
        [ 0, 0, 0, 0, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_t: [
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_u: [
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_v: [
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_w: [
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 1, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 1, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 1, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 1, 0, 1, 0, 0, 1, 0 ],
        [ 1, 0, 1, 0, 0, 0, 1, 0, 1, 0 ],
        [ 1, 1, 0, 0, 0, 0, 0, 1, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_x: [
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 1, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_y: [
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.fat_lcd_z: [
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.truchet_maze_00: [
        [ 1, 0, 0, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 1 ],
    ],

    ValueMask.truchet_maze_01: [
        [ 0, 0, 0, 0, 0, 1 ],
        [ 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.truchet_maze_02: [
        [ 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 0, 1, 1 ],
        [ 1, 1, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0 ],
    ],

    ValueMask.truchet_maze_03: [
        [ 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0 ],
        [ 1, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 1 ],
        [ 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0 ],
    ],

    ValueMask.truchet_tile_00: [
        [ 0, 0, 0, 0, 0, 1 ],
        [ 0, 0, 0, 0, 1, 1 ],
        [ 0, 0, 0, 1, 1, 1 ],
        [ 0, 0, 1, 1, 1, 1 ],
        [ 0, 1, 1, 1, 1, 1 ],
        [ 1, 1, 1, 1, 1, 1 ],
    ],

    ValueMask.truchet_tile_01: [
        [ 1, 1, 1, 1, 1, 1 ],
        [ 0, 1, 1, 1, 1, 1 ],
        [ 0, 0, 1, 1, 1, 1 ],
        [ 0, 0, 0, 1, 1, 1 ],
        [ 0, 0, 0, 0, 1, 1 ],
        [ 0, 0, 0, 0, 0, 1 ],
    ],

    ValueMask.truchet_tile_02: [
        [ 1, 1, 1, 1, 1, 1 ],
        [ 1, 1, 1, 1, 1, 0 ],
        [ 1, 1, 1, 1, 0, 0 ],
        [ 1, 1, 1, 0, 0, 0 ],
        [ 1, 1, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.truchet_tile_03: [
        [ 1, 0, 0, 0, 0, 0 ],
        [ 1, 1, 0, 0, 0, 0 ],
        [ 1, 1, 1, 0, 0, 0 ],
        [ 1, 1, 1, 1, 0, 0 ],
        [ 1, 1, 1, 1, 1, 0 ],
        [ 1, 1, 1, 1, 1, 1 ],
    ],

    ValueMask.mcpaint_00: [
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.mcpaint_01: [
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
    ],

    ValueMask.mcpaint_02: [
        [ 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 1 ],
        [ 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 1 ],
    ],

    ValueMask.mcpaint_03: [
        [ 1, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 1 ],
        [ 1, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 1 ],
    ],

    ValueMask.mcpaint_04: [
        [ 0, 0, 0, 1, 0, 0, 0, 1 ],
        [ 0, 0, 1, 0, 0, 0, 1, 0 ],
        [ 0, 1, 0, 0, 0, 1, 0, 0 ],
        [ 1, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 1 ],
        [ 0, 0, 1, 0, 0, 0, 1, 0 ],
        [ 0, 1, 0, 0, 0, 1, 0, 0 ],
        [ 1, 0, 0, 0, 1, 0, 0, 0 ],
    ],

    ValueMask.mcpaint_05: [
        [ 1, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 1 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.mcpaint_06: [
        [ 1, 1, 0, 0, 1, 1, 0, 0 ],
        [ 0, 1, 1, 0, 0, 1, 1, 0 ],
        [ 0, 0, 1, 1, 0, 0, 1, 1 ],
        [ 1, 0, 0, 1, 1, 0, 0, 1 ],
        [ 1, 1, 0, 0, 1, 1, 0, 0 ],
        [ 0, 1, 1, 0, 0, 1, 1, 0 ],
        [ 0, 0, 1, 1, 0, 0, 1, 1 ],
        [ 1, 0, 0, 1, 1, 0, 0, 1 ],
    ],

    ValueMask.mcpaint_07: [
        [ 1, 0, 0, 1, 1, 0, 0, 1 ],
        [ 0, 0, 1, 1, 0, 0, 1, 1 ],
        [ 0, 1, 1, 0, 0, 1, 1, 0 ],
        [ 1, 1, 0, 0, 1, 1, 0, 0 ],
        [ 1, 0, 0, 1, 1, 0, 0, 1 ],
        [ 0, 0, 1, 1, 0, 0, 1, 1 ],
        [ 0, 1, 1, 0, 0, 1, 1, 0 ],
        [ 1, 1, 0, 0, 1, 1, 0, 0 ],
    ],

    ValueMask.mcpaint_08: [
        [ 1, 0, 1, 0, 1, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 0, 1, 0, 1, 0, 1 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 1, 0, 1, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 0, 1, 0, 1, 0, 1 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.mcpaint_09: [
        [ 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 1, 0, 0, 1, 0, 0 ],
        [ 0, 1, 0, 0, 1, 0, 1, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 1, 0, 1 ],
        [ 0, 1, 0, 1, 0, 0, 1, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 1 ],
        [ 0, 1, 0, 0, 0, 1, 0, 0 ],
    ],

    ValueMask.mcpaint_10: [
        [ 1, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 1 ],
        [ 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 0 ],
    ],

    ValueMask.mcpaint_11: [
        [ 0, 0, 0, 1, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 1, 0 ],
        [ 0, 1, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 1 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 1, 0, 0 ],
        [ 1, 0, 0, 0, 1, 0, 0, 0 ],
    ],

    ValueMask.mcpaint_12: [
        [ 0, 0, 0, 1, 0, 0, 0, 1 ],
        [ 0, 1, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 1 ],
        [ 0, 1, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 1 ],
        [ 0, 1, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 1 ],
        [ 0, 1, 0, 0, 0, 1, 0, 0 ],
    ],

    ValueMask.mcpaint_13: [
        [ 1, 0, 1, 0, 1, 0, 1, 0 ],
        [ 0, 1, 0, 1, 0, 1, 0, 1 ],
        [ 1, 0, 1, 0, 1, 0, 1, 0 ],
        [ 0, 1, 0, 1, 0, 1, 0, 1 ],
        [ 1, 0, 1, 0, 1, 0, 1, 0 ],
        [ 0, 1, 0, 1, 0, 1, 0, 1 ],
        [ 1, 0, 1, 0, 1, 0, 1, 0 ],
        [ 0, 1, 0, 1, 0, 1, 0, 1 ],
    ],

    ValueMask.mcpaint_14: [
        [ 1, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 1 ],
    ],

    ValueMask.mcpaint_15: [
        [ 0, 0, 1, 0, 0, 0, 1, 0 ],
        [ 0, 1, 0, 1, 0, 1, 0, 1 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 1, 0, 1, 0, 1, 0, 1 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.mcpaint_16: [
        [ 0, 0, 1, 0, 0, 0, 1, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 1 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 1 ],
        [ 0, 0, 1, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 1, 1, 1, 0, 0 ],
    ],

    ValueMask.mcpaint_17: [
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
    ],

    ValueMask.mcpaint_18: [
        [ 1, 0, 1, 0, 1, 0, 1, 0 ],
        [ 1, 0, 1, 0, 1, 0, 1, 0 ],
        [ 1, 0, 1, 0, 1, 0, 1, 0 ],
        [ 1, 0, 1, 0, 1, 0, 1, 0 ],
        [ 1, 0, 1, 0, 1, 0, 1, 0 ],
        [ 1, 0, 1, 0, 1, 0, 1, 0 ],
        [ 1, 0, 1, 0, 1, 0, 1, 0 ],
        [ 1, 0, 1, 0, 1, 0, 1, 0 ],
    ],

    ValueMask.mcpaint_19: [
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 1, 0, 1, 1, 0, 0 ],
        [ 1, 0, 0, 1, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 1, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 1, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 0 ],
    ],

    ValueMask.mcpaint_20: [
        [ 0, 1, 0, 0, 0, 1, 0, 0 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
        [ 0, 1, 0, 0, 0, 1, 0, 0 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
        [ 0, 1, 0, 0, 0, 1, 0, 0 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
        [ 0, 1, 0, 0, 0, 1, 0, 0 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
    ],

    ValueMask.mcpaint_21: [
        [ 0, 0, 0, 1, 0, 1, 0, 1 ],
        [ 1, 0, 1, 0, 0, 0, 1, 0 ],
        [ 0, 1, 0, 1, 0, 1, 0, 0 ],
        [ 1, 0, 0, 0, 1, 0, 1, 0 ],
        [ 0, 1, 0, 1, 0, 0, 0, 1 ],
        [ 0, 0, 1, 0, 1, 0, 1, 0 ],
        [ 0, 1, 0, 0, 0, 1, 0, 1 ],
        [ 1, 0, 1, 0, 1, 0, 0, 0 ],
    ],

    ValueMask.mcpaint_22: [
        [ 0, 1, 0, 0, 0, 1, 0, 0 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
        [ 0, 0, 0, 1, 0, 0, 0, 1 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
        [ 0, 1, 0, 0, 0, 1, 0, 0 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
        [ 0, 0, 0, 1, 0, 0, 0, 1 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
    ],

    ValueMask.mcpaint_23: [
        [ 0, 0, 0, 1, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 1, 0, 0, 0 ],
        [ 1, 1, 0, 0, 0, 1, 1, 1 ],
        [ 0, 0, 0, 0, 0, 0, 0, 1 ],
        [ 0, 0, 0, 0, 0, 0, 0, 1 ],
        [ 1, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 1, 1, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 0 ],
    ],

    ValueMask.mcpaint_24: [
        [ 1, 0, 1, 0, 1, 0, 1, 0 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
        [ 1, 0, 1, 0, 1, 0, 1, 0 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
        [ 1, 0, 1, 0, 1, 0, 1, 0 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
        [ 1, 0, 1, 0, 1, 0, 1, 0 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
    ],

    ValueMask.mcpaint_25: [
        [ 1, 0, 0, 0, 1, 0, 0, 0 ],
        [ 1, 0, 0, 0, 1, 0, 0, 0 ],
        [ 1, 0, 0, 0, 1, 0, 0, 0 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
        [ 1, 0, 0, 0, 1, 0, 0, 0 ],
        [ 1, 0, 0, 0, 1, 0, 0, 0 ],
        [ 1, 0, 0, 0, 1, 0, 0, 0 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
    ],

    ValueMask.mcpaint_26: [
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 1, 1, 1, 1, 1, 1 ],
        [ 0, 1, 1, 1, 1, 1, 1, 1 ],
        [ 0, 1, 1, 0, 0, 0, 0, 1 ],
        [ 0, 1, 1, 0, 0, 0, 0, 1 ],
        [ 0, 1, 1, 0, 0, 0, 0, 1 ],
        [ 0, 1, 1, 0, 0, 0, 0, 1 ],
        [ 0, 1, 1, 1, 1, 1, 1, 1 ],
    ],

    ValueMask.mcpaint_27: [
        [ 1, 0, 1, 0, 1, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.mcpaint_28: [
        [ 1, 0, 1, 0, 1, 0, 1, 0 ],
        [ 0, 1, 0, 0, 0, 1, 0, 0 ],
        [ 1, 0, 1, 0, 1, 0, 1, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 1 ],
        [ 1, 0, 1, 0, 1, 0, 1, 0 ],
        [ 0, 1, 0, 0, 0, 1, 0, 0 ],
        [ 1, 0, 1, 0, 1, 0, 1, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 1 ],
    ],

    ValueMask.mcpaint_29: [
        [ 1, 1, 0, 0, 0, 0, 0, 1 ],
        [ 1, 1, 1, 0, 0, 0, 1, 1 ],
        [ 1, 1, 1, 1, 0, 1, 1, 1 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
        [ 1, 1, 1, 1, 0, 1, 1, 1 ],
        [ 1, 1, 1, 0, 0, 0, 1, 1 ],
        [ 1, 1, 0, 0, 0, 0, 0, 1 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.mcpaint_30: [
        [ 0, 1, 0, 0, 0, 0, 0, 1 ],
        [ 1, 0, 1, 0, 0, 0, 1, 0 ],
        [ 0, 1, 0, 1, 0, 1, 0, 1 ],
        [ 1, 0, 1, 0, 1, 0, 1, 0 ],
        [ 0, 1, 0, 1, 0, 1, 0, 1 ],
        [ 1, 0, 1, 0, 0, 0, 1, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 1 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.mcpaint_31: [
        [ 1, 0, 1, 0, 1, 0, 1, 0 ],
        [ 0, 1, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 1 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 0, 1, 0, 1, 0, 1 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 1 ],
    ],

    ValueMask.mcpaint_32: [
        [ 0, 0, 0, 0, 0, 0, 0, 1 ],
        [ 0, 1, 0, 1, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 1 ],
        [ 1, 0, 0, 1, 0, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 1 ],
        [ 0, 1, 0, 1, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 1 ],
        [ 1, 0, 0, 1, 0, 1, 0, 0 ],
    ],

    ValueMask.mcpaint_33: [
        [ 0, 1, 0, 1, 0, 1, 0, 1 ],
        [ 1, 0, 1, 1, 1, 0, 1, 1 ],
        [ 1, 1, 1, 0, 1, 1, 1, 0 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
        [ 1, 0, 1, 0, 1, 0, 1, 0 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
        [ 1, 0, 1, 1, 1, 0, 1, 1 ],
        [ 1, 1, 1, 0, 1, 1, 1, 0 ],
    ],

    ValueMask.mcpaint_34: [
        [ 1, 1, 1, 1, 1, 1, 1, 0 ],
        [ 1, 0, 1, 0, 1, 1, 0, 1 ],
        [ 1, 1, 1, 1, 1, 1, 1, 0 ],
        [ 0, 1, 1, 0, 1, 0, 1, 1 ],
        [ 1, 1, 1, 1, 1, 1, 1, 0 ],
        [ 1, 0, 1, 0, 1, 1, 0, 1 ],
        [ 1, 1, 1, 1, 1, 1, 1, 0 ],
        [ 0, 1, 1, 0, 1, 0, 1, 1 ],
    ],

    ValueMask.mcpaint_35: [
        [ 1, 1, 1, 1, 0, 1, 1, 0 ],
        [ 0, 0, 0, 0, 0, 1, 1, 0 ],
        [ 0, 1, 1, 0, 0, 0, 0, 0 ],
        [ 0, 1, 1, 0, 1, 1, 1, 1 ],
        [ 0, 1, 1, 0, 1, 1, 1, 1 ],
        [ 0, 1, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 1, 1, 0 ],
        [ 1, 1, 1, 1, 0, 1, 1, 0 ],
    ],

    ValueMask.mcpaint_36: [
        [ 0, 0, 0, 1, 1, 1, 0, 0 ],
        [ 0, 1, 0, 0, 1, 1, 1, 0 ],
        [ 1, 1, 1, 0, 0, 1, 0, 0 ],
        [ 1, 1, 0, 0, 0, 0, 0, 1 ],
        [ 1, 0, 0, 0, 0, 0, 1, 1 ],
        [ 0, 0, 1, 0, 0, 1, 1, 1 ],
        [ 0, 1, 1, 1, 0, 0, 1, 0 ],
        [ 0, 0, 1, 1, 1, 0, 0, 0 ],
    ],

    ValueMask.mcpaint_37: [
        [ 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 1, 0, 0, 0, 1, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 0, 1, 1, 1, 1, 1 ],
        [ 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 1, 1, 1, 1, 0, 1, 0, 0 ],
    ],

    ValueMask.mcpaint_38: [
        [ 0, 0, 0, 0, 0, 0, 1, 1 ],
        [ 0, 1, 1, 0, 0, 1, 1, 0 ],
        [ 0, 0, 1, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 1, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 1, 0, 0 ],
        [ 0, 1, 1, 0, 0, 1, 1, 0 ],
        [ 1, 1, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 1 ],
    ],

    ValueMask.mcpaint_39: [
        [ 1, 0, 0, 0, 1, 0, 0, 0 ],
        [ 1, 1, 0, 1, 1, 1, 0, 1 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
        [ 1, 1, 0, 1, 1, 1, 0, 1 ],
        [ 1, 0, 0, 0, 1, 0, 0, 0 ],
        [ 1, 1, 0, 1, 1, 1, 0, 1 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1 ],
        [ 1, 1, 0, 1, 1, 1, 0, 1 ],
    ],

    ValueMask.mcpaint_40: [
        [ 1, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 1, 0, 1, 0, 0, 0, 0 ],
        [ 1, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 1, 0, 1, 0, 0, 0, 0 ],
        [ 1, 1, 1, 1, 1, 0, 1, 0 ],
        [ 1, 1, 1, 1, 0, 1, 0, 1 ],
        [ 1, 1, 1, 1, 1, 0, 1, 0 ],
        [ 1, 1, 1, 1, 0, 1, 0, 1 ],
    ],

    ValueMask.emoji_00: [
        [ 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0 ],
        [ 1, 1, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 1, 1, 0 ],
        [ 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.emoji_01: [
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 1, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0 ],
        [ 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.emoji_02: [
        [ 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0 ],
        [ 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0 ],
        [ 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0 ],
        [ 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0 ],
        [ 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0 ],
        [ 0, 1, 0, 0, 1, 0, 1, 0, 1, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.emoji_03: [
        [ 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.emoji_04: [
        [ 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0 ],
        [ 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.emoji_05: [
        [ 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0 ],
        [ 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0 ],
        [ 0, 1, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0 ],
        [ 0, 1, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0 ],
        [ 0, 1, 0, 0, 1, 1, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.emoji_06: [
        [ 1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 1, 1, 1, 0, 1, 1, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 1, 1, 1, 0, 1, 1, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.emoji_07: [
        [ 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.emoji_08: [
        [ 0, 0, 1, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0 ],
        [ 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0 ],
        [ 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 1, 0 ],
        [ 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.emoji_09: [
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.emoji_10: [
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0 ],
        [ 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.emoji_11: [
        [ 0, 0, 1, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0 ],
        [ 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0 ],
        [ 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 1, 0 ],
        [ 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.emoji_12: [
        [ 1, 1, 0, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0 ],
        [ 0, 1, 0, 0, 1, 0, 1, 0, 1, 0, 0, 1, 0 ],
        [ 0, 1, 0, 1, 0, 0, 0, 1, 1, 0, 0, 1, 0 ],
        [ 0, 1, 0, 1, 0, 0, 0, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.emoji_13: [
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0 ],
        [ 0, 1, 1, 0, 1, 1, 0, 1, 1, 0, 1, 0, 0 ],
        [ 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0 ],
        [ 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0 ],
        [ 0, 1, 1, 0, 1, 1, 0, 1, 1, 0, 1, 0, 0 ],
        [ 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0 ],
        [ 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.emoji_14: [
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.emoji_15: [
        [ 0, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 1, 1, 1, 0, 1, 0, 0, 0, 0 ],
        [ 0, 0, 1, 1, 1, 1, 0, 0, 1, 1, 0, 0, 0 ],
        [ 0, 0, 1, 1, 1, 0, 0, 1, 1, 1, 0, 0, 0 ],
        [ 0, 1, 1, 1, 0, 0, 1, 1, 1, 1, 1, 0, 0 ],
        [ 0, 1, 1, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0 ],
        [ 0, 1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0 ],
        [ 0, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0 ],
        [ 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.emoji_16: [
        [ 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 1, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0 ],
        [ 0, 1, 0, 1, 0, 0, 0, 0, 1, 0, 1, 0, 0 ],
        [ 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0 ],
        [ 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0 ],
        [ 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0 ],
        [ 0, 1, 0, 1, 0, 0, 0, 0, 1, 0, 1, 0, 0 ],
        [ 0, 0, 1, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.emoji_17: [
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0 ],
        [ 0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0 ],
        [ 0, 1, 1, 1, 1, 1, 1, 0, 1, 0, 0, 0, 0 ],
        [ 0, 1, 1, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0 ],
        [ 0, 1, 1, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.emoji_18: [
        [ 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 1, 0, 1, 1, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0 ],
        [ 0, 0, 1, 1, 0, 1, 0, 1, 0, 1, 0, 0, 0 ],
        [ 0, 0, 1, 0, 1, 0, 1, 0, 1, 1, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 1, 1, 0, 1, 0, 1, 1, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.emoji_19: [
        [ 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 1, 1, 1, 1, 1, 0, 1, 1, 0, 0, 0 ],
        [ 0, 1, 0, 0, 1, 1, 1, 0, 0, 1, 1, 0, 0 ],
        [ 0, 1, 1, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0 ],
        [ 1, 1, 0, 1, 0, 1, 0, 0, 0, 1, 0, 1, 0 ],
        [ 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0 ],
        [ 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 1, 0 ],
        [ 0, 1, 1, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0 ],
        [ 0, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 1, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.emoji_20: [
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0 ],
        [ 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.emoji_21: [
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0 ],
        [ 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 1, 0, 0 ],
        [ 0, 1, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 0 ],
        [ 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 1, 0, 0 ],
        [ 0, 0, 1, 0, 1, 0, 1, 0, 1, 1, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.emoji_22: [
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 0 ],
        [ 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 0 ],
        [ 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0 ],
        [ 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0 ],
        [ 0, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 0 ],
        [ 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.emoji_23: [
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0 ],
        [ 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 0 ],
        [ 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0 ],
        [ 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0 ],
        [ 0, 1, 1, 1, 0, 1, 1, 0, 1, 0, 1, 1, 0 ],
        [ 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.emoji_24: [
        [ 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0 ],
        [ 0, 0, 1, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0 ],
        [ 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0 ],
        [ 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0 ],
        [ 0, 0, 1, 1, 0, 1, 1, 0, 1, 1, 0, 0, 0 ],
        [ 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.emoji_25: [
        [ 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0 ],
        [ 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0 ],
        [ 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 0 ],
        [ 1, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 1, 0 ],
        [ 1, 1, 0, 0, 1, 1, 1, 1, 0, 0, 1, 1, 0 ],
        [ 1, 1, 0, 0, 0, 1, 1, 0, 0, 0, 1, 1, 0 ],
        [ 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0 ],
        [ 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0 ],
        [ 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0 ],
        [ 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0 ],
        [ 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0 ],
        [ 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.emoji_26: [
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0 ],
        [ 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0 ],
        [ 0, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.emoji_27: [
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.bank_ocr_1: [
        [ 0, 0, 1, 1, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.bank_ocr_2: [
        [ 0, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0 ],
        [ 0, 0, 1, 0, 0, 0, 0 ],
        [ 0, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.bank_ocr_3: [
        [ 0, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 1, 1, 0 ],
        [ 0, 0, 0, 0, 1, 1, 0 ],
        [ 0, 0, 1, 1, 1, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.bank_ocr_4: [
        [ 1, 1, 0, 0, 0, 0, 0 ],
        [ 1, 1, 0, 0, 0, 0, 0 ],
        [ 1, 1, 0, 0, 0, 0, 0 ],
        [ 1, 1, 0, 0, 0, 0, 0 ],
        [ 1, 1, 0, 0, 1, 1, 0 ],
        [ 1, 1, 1, 1, 1, 1, 0 ],
        [ 0, 0, 0, 0, 1, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.bank_ocr_5: [
        [ 0, 1, 1, 1, 1, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 1, 0, 0, 0, 0, 0 ],
        [ 0, 1, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 1, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.bank_ocr_6: [
        [ 1, 1, 1, 1, 0, 0, 0 ],
        [ 1, 0, 0, 1, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0 ],
        [ 1, 0, 0, 0, 0, 0, 0 ],
        [ 1, 1, 1, 1, 1, 1, 0 ],
        [ 1, 0, 0, 0, 0, 1, 0 ],
        [ 1, 1, 1, 1, 1, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.bank_ocr_7: [
        [ 0, 1, 1, 1, 1, 0, 0 ],
        [ 0, 1, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 0, 1, 0, 0 ],
        [ 0, 0, 0, 1, 1, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 1, 0, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.bank_ocr_8: [
        [ 0, 1, 1, 1, 1, 0, 0 ],
        [ 0, 1, 0, 0, 1, 0, 0 ],
        [ 0, 1, 0, 0, 1, 0, 0 ],
        [ 1, 1, 1, 1, 1, 1, 0 ],
        [ 1, 1, 0, 0, 1, 1, 0 ],
        [ 1, 1, 0, 0, 1, 1, 0 ],
        [ 1, 1, 1, 1, 1, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.bank_ocr_9: [
        [ 0, 1, 1, 1, 1, 1, 0 ],
        [ 0, 1, 0, 0, 0, 1, 0 ],
        [ 0, 1, 0, 0, 0, 1, 0 ],
        [ 0, 1, 1, 1, 1, 1, 0 ],
        [ 0, 0, 0, 0, 1, 1, 0 ],
        [ 0, 0, 0, 0, 1, 1, 0 ],
        [ 0, 0, 0, 0, 1, 1, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.bank_ocr_0: [
        [ 0, 1, 1, 1, 1, 0, 0 ],
        [ 1, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 1, 0 ],
        [ 1, 0, 0, 0, 0, 1, 0 ],
        [ 0, 1, 1, 1, 1, 0, 0 ],
        [ 0, 0, 0, 0, 0, 0, 0 ],
    ],

    ValueMask.conv2d_blur: [
        [1, 4, 6, 4, 1],
        [4, 16, 24, 16, 4],
        [6, 24, 36, 24, 6],
        [4, 16, 24, 16, 4],
        [1, 4, 6, 4, 1]
    ],

    ValueMask.conv2d_deriv_x: [
        [0, 0, 0],
        [0, 1, -1],
        [0, 0, 0]
    ],

    ValueMask.conv2d_deriv_y: [
        [0, 0, 0],
        [0, 1, 0],
        [0, -1, 0]
    ],

    ValueMask.conv2d_edges: [
        [1, 2, 1],
        [2, -12, 2],
        [1, 2, 1]
    ],

    ValueMask.conv2d_emboss: [
        [0, 2, 4],
        [-2, 1, 2],
        [-4, -2, 0]
    ],

    ValueMask.conv2d_invert: [
        [0, 0, 0],
        [0, -1, 0],
        [0, 0, 0]
    ],

    ValueMask.conv2d_rand: np.random.normal(.5, .5, (5, 5)).tolist(),

    ValueMask.conv2d_sharpen: [
        [0, -1, 0],
        [-1, 5, -1],
        [0, -1, 0]
    ],

    ValueMask.conv2d_sobel_x: [
        [1, 0, -1],
        [2, 0, -2],
        [1, 0, -1]
    ],

    ValueMask.conv2d_sobel_y: [
        [1, 2, 1],
        [0, 0, 0],
        [-1, -2, -1]
    ],

    ValueMask.conv2d_box_blur: [
        [1, 2, 1],
        [2, 4, 2],
        [1, 2, 1]
    ],

    ValueMask.rgb: [
        [[1, 0, 0], [0, 1, 0], [0, 0, 1], [0, 0, 0]],
        [[1, 0, 0], [0, 1, 0], [0, 0, 1], [0, 0, 0]],
        [[1, 0, 0], [0, 1, 0], [0, 0, 1], [0, 0, 0]],
        [[0, 0, 0], [0, 0, 0], [0, 0, 0], [0, 0, 0]]
    ],

    ValueMask.rbggbr: [
        [[1, 0, 0], [0, 0, 1], [0, 1, 0], [1, 0, 0], [0, 0, 1], [0, 1, 0]],
        [[1, 0, 0], [0, 0, 1], [0, 1, 0], [1, 0, 0], [0, 0, 1], [0, 1, 0]],
        [[1, 0, 0], [0, 0, 1], [0, 1, 0], [1, 0, 0], [0, 0, 1], [0, 1, 0]],
        [[0, 1, 0], [0, 0, 1], [1, 0, 0], [0, 1, 0], [0, 0, 1], [1, 0, 0]],
        [[0, 1, 0], [0, 0, 1], [1, 0, 0], [0, 1, 0], [0, 0, 1], [1, 0, 0]],
        [[0, 1, 0], [0, 0, 1], [1, 0, 0], [0, 1, 0], [0, 0, 1], [1, 0, 0]],
    ],

    ValueMask.rgbgr: [
        [[1, 0, 0], [1, 0, 0], [1, 0, 0], [0, 0, 1], [0, 1, 0], [0, 1, 0], [0, 1, 0]],
        [[1, 0, 0], [1, 0, 0], [0, 0, 1], [0, 0, 1], [0, 0, 1], [0, 1, 0], [0, 1, 0]],
        [[1, 0, 0], [0, 0, 1], [0, 0, 1], [0, 0, 1], [0, 0, 1], [0, 0, 1], [0, 1, 0]],
        [[0, 0, 1], [0, 0, 1], [0, 0, 1], [0, 0, 1], [0, 0, 1], [0, 0, 1], [0, 0, 1]],
        [[0, 1, 0], [0, 0, 1], [0, 0, 1], [0, 0, 1], [0, 0, 1], [0, 0, 1], [1, 0, 0]],
        [[0, 1, 0], [0, 1, 0], [0, 0, 1], [0, 0, 1], [0, 0, 1], [1, 0, 0], [1, 0, 0]],
        [[0, 1, 0], [0, 1, 0], [0, 1, 0], [0, 0, 1], [1, 0, 0], [1, 0, 0], [1, 0, 0]],
    ],

    ValueMask.roygbiv: [
        [[1, 0, .5], [1, .5, 0], [1, 1, 0], [.5, 1, 0], [0, .5, 1], [.5, 0, 1], [.75, 0, .75], [0, 0, 0]],
        [[1, 0, .5], [1, .5, 0], [1, 1, 0], [.5, 1, 0], [0, .5, 1], [.5, 0, 1], [.75, 0, .75], [0, 0, 0]],
        [[1, 0, .5], [1, .5, 0], [1, 1, 0], [.5, 1, 0], [0, .5, 1], [.5, 0, 1], [.75, 0, .75], [0, 0, 0]],
        [[1, 0, .5], [1, .5, 0], [1, 1, 0], [.5, 1, 0], [0, .5, 1], [.5, 0, 1], [.75, 0, .75], [0, 0, 0]],
        [[1, 0, .5], [1, .5, 0], [1, 1, 0], [.5, 1, 0], [0, .5, 1], [.5, 0, 1], [.75, 0, .75], [0, 0, 0]],
        [[1, 0, .5], [1, .5, 0], [1, 1, 0], [.5, 1, 0], [0, .5, 1], [.5, 0, 1], [.75, 0, .75], [0, 0, 0]],
        [[1, 0, .5], [1, .5, 0], [1, 1, 0], [.5, 1, 0], [0, .5, 1], [.5, 0, 1], [.75, 0, .75], [0, 0, 0]],
        [[0, 0, 0], [0, 0, 0], [0, 0, 0], [0, 0, 0], [0, 0, 0], [0, 0, 0], [0, 0, 0], [0, 0, 0]]
    ],

    ValueMask.rggb: [
        [[1, 0, 0], [0, 1, 0]],
        [[0, 1, 0], [0, 0, 1]],
    ],

    ValueMask.rainbow: [
        [[1, 0, .5]] * 7,
        [[1, .5, 0]] * 7,
        [[1, 1, 0]] * 7,
        [[.5, 1, 0]] * 7,
        [[0, .5, 1]] * 7,
        [[.5, 0, 1]] * 7,
        [[.75, 0, .75]] * 7,
    ],

    ValueMask.ace: [
        [[0, 0, 0]] * 4,
        [[.5, .5, .5]] * 4,
        [[1, 1, 1]] * 4,
        [[.625, 0, .625]] * 4,
    ],

    ValueMask.nb: [
        [[1, 1, 0]] * 4,
        [[1, 1, 1]] * 4,
        [[.625, 0, .625]] * 4,
        [[0, 0, 0]] * 4,
    ],

    ValueMask.trans: [
        [[0, 1, 1]] * 5 + [[0, 0, 0]],
        [[1, .5, .75]] * 5 + [[0, 0, 0]],
        [[1, 1, 1]] * 5 + [[0, 0, 0]],
        [[1, .5, .75]] * 5 + [[0, 0, 0]],
        [[0, 1, 1]] * 5 + [[0, 0, 0]],
        [[0, 0, 0]] * 6,
    ],
}

# Procedural masks, corresponding to keys in constants.ValueMask

# Procedural mask shapes

_mask_shapes = {}


def mask(*args):
    """Function decorator for procedural masks."""

    def decorator_fn(mask_fn):
        mask = ValueMask[mask_fn.__name__]

        if args:
            _mask_shapes[mask] = args[0]

        Masks[mask] = mask_fn

        return mask_fn

    return decorator_fn


def mask_shape(mask):
    """Return the shape for the received ValueMask.

    :param ValueMask mask:
    """

    if mask in _mask_shapes:
        shape = _mask_shapes[mask]

        if callable(shape):
            shape = shape()

    else:
        height = len(Masks[mask])
        width = len(Masks[mask][0])

        if isinstance(Masks[mask][0][0], list):
            channels = len(Masks[mask][0][0])
        else:
            channels = 1

        shape = [height, width, channels]

    return shape


def mask_values(mask, channel_shape=None, uv_noise=None, atlas=None, inverse=False, time=0.0, speed=1.0):
    """
    Return a tuple of (pixel values, brightness) for the received ValueMask.

    :param ValueMask mask:
    :param list[int] channel_shape:
    :param None|Tensor uv_noise:
    :param None|list atlas:
    :param bool inverse:
    """

    shape = mask_shape(mask)

    if channel_shape is None:
        channel_shape = shape

    if len(shape) == 3:
        channel_shape[2] = shape[2]

    mask_values = []

    uv_shape = [int(channel_shape[0] / shape[0]) or 1, int(channel_shape[1] / shape[1]) or 1]

    if uv_noise is None:
        uv_noise = simplex(uv_shape, time=time, seed=random.randint(1, 65536), speed=speed, as_np=True)

        # normalize() but it's numpy
        floor = np.amin(uv_noise)
        ceil = np.amax(uv_noise)
        uv_noise = (uv_noise - floor) / (ceil - floor)

    total = 0

    for y in range(channel_shape[0]):
        uv_y = int((y / channel_shape[0]) * uv_shape[0])

        mask_row = []
        mask_values.append(mask_row)

        for x in range(channel_shape[1]):
            uv_x = int((x / channel_shape[1]) * uv_shape[1])

            if callable(Masks[mask]):
                pixel = Masks[mask](x=x, y=y, row=mask_row, shape=shape, uv_x=uv_x, uv_y=uv_y, uv_noise=uv_noise, uv_shape=uv_shape, atlas=atlas, channel_shape=channel_shape)

            else:
                pixel = Masks[mask][y % shape[0]][x % shape[1]]

            if not isinstance(pixel, list):
                pixel = [pixel]

            pixel = [float(i) for i in pixel]

            if inverse:
                pixel = [1.0 - i for i in pixel]

            mask_row.append(pixel)

            total += sum(pixel)

    return mask_values, total


def square_masks():
    """Return a list of square ValueMasks."""

    square = []

    for mask in ValueMask:
        if callable(_mask_shapes.get(mask)):  # No dynamic shapes
            continue

        shape = mask_shape(mask)

        if shape and shape[0] == shape[1]:
            square.append(mask)

    return square


def _glyph_from_atlas_range(x, y, shape, uv_x, uv_y, uv_noise, min_value, max_value):
    atlas = [Masks[g] for g in Masks if g.value >= min_value and g.value <= max_value]

    glyph_index = int(uv_noise[uv_y][uv_x] * (len(atlas)))

    glyph_index = min(max(glyph_index, 0), len(atlas) - 1)

    return atlas[glyph_index][y % shape[0]][x % shape[1]]


@mask([10, 10, 1])
def sparse(**kwargs):
    return 1 if random.random() < .15 else 0


@mask([10, 10, 1])
def sparser(**kwargs):
    return 1 if random.random() < .05 else 0


@mask(lambda: [random.randint(5, 7), random.randint(6, 12), 1])
def invaders(**kwargs):
    return _invaders(**kwargs)


@mask([6, 6, 1])
def invaders_square(**kwargs):
    return _invaders(**kwargs)


@mask([4, 4, 1])
def white_bear(**kwargs):
    return _invaders(**kwargs)


def _invaders(x, y, row, shape, **kwargs):
    # Inspired by http://www.complexification.net/gallery/machines/invaderfractal/
    height = shape[0]
    width = shape[1]

    if y % height == 0 or x % width == 0:
        return 0

    elif x % width > width / 2:
        return row[x - int(((x % width) - width / 2) * 2)]

    else:
        return random.randint(0, 1)


@mask([6, 4, 1])
def matrix(x, y, row, shape, **kwargs):
    height = shape[0]
    width = shape[1]

    if y % height == 0 or x % width == 0:
        return 0

    return random.randint(0, 1)


@mask(lambda: [random.randint(3, 4) * 2 + 1, random.randint(3, 4) * 2 + 1, 1])
def letters(x, y, row, shape, **kwargs):
    # Inspired by https://www.shadertoy.com/view/4lscz8
    height = shape[0]
    width = shape[1]

    if any(n == 0 for n in (x % width, y % height)):
        return 0

    if any(n == 1 for n in (width - (x % width), height - (y % height))):
        return 0

    if all(n % 2 == 0 for n in (x % width, y % height)):
        return 0

    if x % 2 == 0 or y % 2 == 0:
        return random.random() > .25

    return random.random() > .75


@mask([14, 8, 1])
def iching(x, y, row, shape, **kwargs):
    height = shape[0]
    width = shape[1]

    if any(n == 0 for n in (x % width, y % height)):
        return 0

    if any(n == 1 for n in (width - (x % width), height - (y % height))):
        return 0

    if y % 2 == 0:
        return 0

    if x % 2 == 1 and x % width not in (3, 4):
        return 1

    if x % 2 == 0:
        return row[x - 1]

    return random.randint(0, 1)


@mask(lambda: [random.randint(4, 6) * 2] * 2 + [1])
def ideogram(x, y, row, shape, **kwargs):
    height = shape[0]
    width = shape[1]

    if any(n == 0 for n in (x % width, y % height)):
        return 0

    if any(n == 1 for n in (width - (x % width), height - (y % height))):
        return 0

    if all(n % 2 == 1 for n in (x % width, y % height)):
        return 0

    return random.random() > .5


@mask(lambda: [random.randint(7, 9), random.randint(12, 24), 1])
def script(x, y, row, shape, **kwargs):
    height = shape[0]
    width = shape[1]

    x_step = x % width
    y_step = y % height

    if x > 0 and (x + y) % 2 == 1:
        return row[x - 1]

    if y_step == 0:
        return 0

    if y_step in (1, 3, 6):
        return random.random() > .25

    if y_step in (2, 4, 5):
        return random.random() > .9

    if x_step == 0:
        return 0

    if any(n == 0 for n in (width - x_step, height - y_step)):
        return 0

    if all(n % 2 == 0 for n in (x_step, y_step)):
        return 0

    if y_step == height - 1:
        return 0

    return random.random() > .5


@mask([6, 6, 1])
def binary(x, y, row, shape, uv_x, uv_y, uv_noise, **kwargs):
    glyph = Masks[ValueMask.zero] if uv_noise[uv_y][uv_x] < .5 else Masks[ValueMask.one]

    return glyph[y % shape[0]][x % shape[1]]


@mask([4, 4, 1])
def tromino(x, y, row, shape, uv_x, uv_y, uv_noise, uv_shape, **kwargs):
    atlas = [Masks[g] for g in Masks if g.name.startswith("tromino_")]

    tex_x = x % shape[1]
    tex_y = y % shape[0]

    uv_value = uv_noise[uv_y][uv_x] * (len(atlas) - 1)
    uv_floor = int(uv_value)
    uv_fract = uv_value - uv_floor

    float2 = uv_noise[(uv_y + int(shape[0] * .5)) % uv_shape[0]][uv_x]
    float3 = uv_noise[uv_y][(uv_x + int(shape[1] * .5)) % uv_shape[1]]

    if uv_fract < .5:
        _x = tex_x
        tex_x = tex_y
        tex_y = _x

    if float2 < .5:
        tex_x = shape[1] - tex_x - 1

    if float3 < .5:
        tex_y = shape[0] - tex_y - 1

    return atlas[uv_floor][tex_x][tex_y]


@mask([6, 6, 1])
def numeric(x, y, row, shape, uv_x, uv_y, uv_noise, **kwargs):
    return _glyph_from_atlas_range(x, y, shape, uv_x, uv_y, uv_noise, ValueMask.zero.value, ValueMask.nine.value)


@mask([6, 6, 1])
def hex(x, y, row, shape, uv_x, uv_y, uv_noise, **kwargs):
    return _glyph_from_atlas_range(x, y, shape, uv_x, uv_y, uv_noise, ValueMask.zero.value, ValueMask.f.value)


@mask([15, 15, 1])
def truetype(x, y, row, shape, uv_x, uv_y, uv_noise, atlas, **kwargs):
    value = max(0, min(1, uv_noise[uv_y][uv_x]))

    glyph = atlas[int(value * (len(atlas) - 1))]

    return glyph[y % shape[0]][x % shape[1]]


@mask([4, 4, 1])
def halftone(x, y, row, shape, uv_x, uv_y, uv_noise, **kwargs):
    return _glyph_from_atlas_range(x, y, shape, uv_x, uv_y, uv_noise, ValueMask.halftone_0.value, ValueMask.halftone_9.value)


@mask([8, 5, 1])
def lcd(x, y, row, shape, uv_x, uv_y, uv_noise, **kwargs):
    return _glyph_from_atlas_range(x, y, shape, uv_x, uv_y, uv_noise, ValueMask.lcd_0.value, ValueMask.lcd_9.value)


@mask([8, 5, 1])
def lcd_binary(x, y, row, shape, uv_x, uv_y, uv_noise, **kwargs):
    return _glyph_from_atlas_range(x, y, shape, uv_x, uv_y, uv_noise, ValueMask.lcd_0.value, ValueMask.lcd_1.value)


@mask([10, 10, 1])
def fat_lcd(x, y, row, shape, uv_x, uv_y, uv_noise, **kwargs):
    return _glyph_from_atlas_range(x, y, shape, uv_x, uv_y, uv_noise, ValueMask.fat_lcd_0.value, ValueMask.fat_lcd_z.value)


@mask([10, 10, 1])
def fat_lcd_binary(x, y, row, shape, uv_x, uv_y, uv_noise, **kwargs):
    return _glyph_from_atlas_range(x, y, shape, uv_x, uv_y, uv_noise, ValueMask.fat_lcd_0.value, ValueMask.fat_lcd_1.value)


@mask([10, 10, 1])
def fat_lcd_numeric(x, y, row, shape, uv_x, uv_y, uv_noise, **kwargs):
    return _glyph_from_atlas_range(x, y, shape, uv_x, uv_y, uv_noise, ValueMask.fat_lcd_0.value, ValueMask.fat_lcd_9.value)


@mask([10, 10, 1])
def fat_lcd_hex(x, y, row, shape, uv_x, uv_y, uv_noise, **kwargs):
    return _glyph_from_atlas_range(x, y, shape, uv_x, uv_y, uv_noise, ValueMask.fat_lcd_0.value, ValueMask.fat_lcd_f.value)


@mask([6, 3, 1])
def arecibo_num(x, y, row, shape, uv_x, uv_y, uv_noise, **kwargs):
    tex_x = x % shape[1]
    tex_y = y % shape[0]

    if tex_y == 0 or tex_y == shape[0] - 1 or tex_x == 0:
        return 0

    if tex_y == shape[0] - 2:
        return 1 if tex_x == 1 else 0

    return random.randint(0, 1)


@mask([6, 5, 1])
def arecibo_bignum(*args, **kwargs):
    return arecibo_num(*args, **kwargs)


@mask([6, 6, 1])
def arecibo_nucleotide(x, y, row, shape, uv_x, uv_y, uv_noise, **kwargs):
    tex_x = x % shape[1]
    tex_y = y % shape[0]

    if tex_y == 0 or tex_y == shape[0] - 1 or tex_x == 0:
        return 0

    if tex_y == shape[0] - 2:
        return 1 if tex_x < shape[1] else 0

    if tex_y < shape[0] - 3 and tex_x > shape[1] - 2:
        return 0

    return random.randint(0, 1)


_ARECIBO_DNA_TEMPLATE = [
    [ 0, 1, 0, 0, 0, 0, 0, 0, -1, -1, 0, 0, 0, 0, 0, 0, 1, 0 ],
    [ 0, 0, 1, 0, 0, 0, 0, 0, -1, -1, 0, 0, 0, 0, 0, 1, 0, 0 ],
    [ 0, 0, 0, 1, 0, 0, 0, 0, -1, -1, 0, 0, 0, 0, 1, 0, 0, 0 ],
    [ 0, 0, 0, 0, 1, 1, 0, 0, -1,  0, 0, 0, 1, 1, 0, 0, 0, 0 ],
    [ 0, 0, 0, 0, 0, 0, 1, 0,  0,  0, 1, 1, 0, 0, 0, 0, 0, 0 ],
    [ 0, 0, 0, 0, 0, 0, 0, 0,  1,  1, 0, 0, 0, 0, 0, 0, 0, 0 ],
    [ 0, 0, 0, 0, 0, 0, 1, 1,  0,  0, 0, 1, 0, 0, 0, 0, 0, 0 ],
    [ 0, 0, 0, 0, 1, 1, 0, 0,  0, -1, 0, 0, 1, 1, 0, 0, 0, 0 ],
    [ 0, 0, 0, 1, 0, 0, 0, 0, -1, -1, 0, 0, 0, 0, 1, 0, 0, 0 ],
    [ 0, 0, 1, 0, 0, 0, 0, 0, -1, -1, 0, 0, 0, 0, 0, 1, 0, 0 ],
    [ 0, 1, 0, 0, 0, 0, 0, 0, -1, -1, 0, 0, 0, 0, 0, 0, 1, 0 ],
]


@mask([11, 17, 1])
def arecibo_dna(x, y, row, shape, uv_x, uv_y, uv_noise, **kwargs):
    tex_x = x % shape[1]
    tex_y = y % shape[0]

    value = _ARECIBO_DNA_TEMPLATE[tex_y][tex_x]

    return random.randint(0, 1) if value == -1 else value


@mask(lambda: [64, 64, 1])
def arecibo(x, y, row, shape, uv_x, uv_y, uv_noise, channel_shape, **kwargs):
    third_height = channel_shape[0] / 3
    half_width = channel_shape[1] / 2
    dna_half_width = mask_shape(ValueMask.arecibo_dna)[1] * .5

    if x > half_width - dna_half_width and x < half_width + dna_half_width:
        dna_x = int(x - half_width - dna_half_width)

        return arecibo_dna(dna_x, y, row, mask_shape(ValueMask.arecibo_dna), uv_x, uv_y, uv_noise, **kwargs)

    if x > half_width - (dna_half_width + 2) and x < half_width + dna_half_width + 1:
        return 0

    if y < third_height:
        return arecibo_num(x, y, row, mask_shape(ValueMask.arecibo_num), uv_x, uv_y, uv_noise, **kwargs)

    if y < third_height * 2:
        return arecibo_nucleotide(x, y, row, mask_shape(ValueMask.arecibo_nucleotide), uv_x, uv_y, uv_noise, **kwargs)

    return arecibo_bignum(x, y, row, mask_shape(ValueMask.arecibo_bignum), uv_x, uv_y, uv_noise, **kwargs)


@mask([6, 6, 1])
def truchet_maze(x, y, row, shape, uv_x, uv_y, uv_noise, **kwargs):
    masks = [
        [ValueMask.truchet_maze_00.value, ValueMask.truchet_maze_01.value],
        [ValueMask.truchet_maze_02.value, ValueMask.truchet_maze_03.value]
    ][0 if uv_noise[0][0] < .5 else 1]

    return _glyph_from_atlas_range(x, y, shape, uv_x, uv_y, uv_noise, *masks)


@mask([6, 6, 1])
def truchet_tile(x, y, row, shape, uv_x, uv_y, uv_noise, **kwargs):
    return _glyph_from_atlas_range(x, y, shape, uv_x, uv_y, uv_noise, ValueMask.truchet_tile_00.value, ValueMask.truchet_tile_03.value)


@mask([8, 8, 1])
def mcpaint(x, y, row, shape, uv_x, uv_y, uv_noise, **kwargs):
    return _glyph_from_atlas_range(x, y, shape, uv_x, uv_y, uv_noise, ValueMask.mcpaint_00.value, ValueMask.mcpaint_40.value)


@mask([13, 13, 1])
def emoji(x, y, row, shape, uv_x, uv_y, uv_noise, **kwargs):
    return _glyph_from_atlas_range(x, y, shape, uv_x, uv_y, uv_noise, ValueMask.emoji_00.value, ValueMask.emoji_26.value)


@mask([24, 1, 1])
def bar_code(x, y, row, shape, uv_x, uv_y, uv_noise, **kwargs):
    return uv_noise[0][uv_x] < .5


@mask([10, 1, 1])
def bar_code_short(*args, **kwargs):
    return bar_code(*args, **kwargs)


@mask([8, 7, 1])
def bank_ocr(x, y, row, shape, uv_x, uv_y, uv_noise, **kwargs):
    return _glyph_from_atlas_range(x, y, shape, uv_x, uv_y, uv_noise, ValueMask.bank_ocr_0.value, ValueMask.bank_ocr_9.value)


@mask(lambda: [random.randint(25, 50)] * 2 + [1])
def fake_qr(x, y, row, shape, uv_x, uv_y, uv_noise, **kwargs):
    x = x % shape[1]
    y = y % shape[1]

    if (x == 0 or y == 0 or x == shape[1] - 1 or y == shape[0] - 1) \
            or (y in (8, shape[0] - 9) and x < 9) \
            or (y == 8 and x > shape[1] - 10) \
            or (x in (8, shape[1] - 9) and y < 9) \
            or (x == 8 and y > shape[0] - 10) \
            or (y in (2, 6) and (x in range(2, 7) or x in range(shape[1] - 7, shape[1] - 2))) \
            or (y in (shape[1] - 3, shape[1] - 7) and x in range(2, 7)) \
            or (x in (2, 6) and (y in range(2, 7) or y in range(shape[0] - 7, shape[0] - 2))) \
            or (x in (shape[0] - 3, shape[0] - 7) and y in range(2, 7)) \
            or (x in (shape[0] - 7, shape[0] - 9) and y in range(shape[0] - 9, shape[0] - 6)) \
            or (y in (shape[1] - 7, shape[1] - 9) and x in range(shape[1] - 9, shape[1] - 6)):

        return 1

    elif (x == shape[1] - 8 and y == shape[0] - 8) \
            or (x in (shape[0] - 6, shape[0] - 10) and y in range(shape[0] - 10, shape[0] - 5)) \
            or (y in (shape[1] - 6, shape[1] - 10) and x in range(shape[1] - 10, shape[1] - 5)):

        return 0

    elif (x > 8 and x < shape[1] - 8) \
            or (y > 8 and y < shape[0] - 8) \
            or (x >= shape[1] - 8 and y >= shape[0] - 8):

        return random.randint(0, 1)

    return 0
