"""
Copyright (c) 2015, 2018, 2019 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.
"""
from __future__ import print_function, unicode_literals, absolute_import

import koji
import os

from atomic_reactor.plugins.pre_koji import KojiPlugin
from atomic_reactor.plugins.pre_reactor_config import (ReactorConfigPlugin,
                                                       WORKSPACE_CONF_KEY,
                                                       ReactorConfig)
from atomic_reactor.core import DockerTasker
from atomic_reactor.inner import DockerBuildWorkflow
from atomic_reactor.plugin import PreBuildPluginsRunner
from flexmock import flexmock
from fnmatch import fnmatch
import pytest
from tests.constants import SOURCE, MOCK
from tests.stubs import StubInsideBuilder, StubSource
if MOCK:
    from tests.docker_mock import mock_docker


KOJI_TARGET = "target"
KOJI_TARGET_BROKEN_TAG = "target-broken"
KOJI_TARGET_BROKEN_REPO = "target-broken-repo"
KOJI_TAG = "tag"
KOJI_BROKEN_TAG = "tag-broken"
KOJI_BROKEN_REPO = "tag-broken-repo"
GET_TARGET_RESPONSE = {"build_tag_name": KOJI_TAG}
BROKEN_TAG_RESPONSE = {"build_tag_name": KOJI_BROKEN_TAG}
BROKEN_REPO_RESPONSE = {"build_tag_name": KOJI_BROKEN_REPO}
TAG_ID = "1"
BROKEN_REPO_TAG_ID = "2"
GET_TAG_RESPONSE = {"id": TAG_ID, "name": KOJI_TAG}
REPO_ID = "2"
BROKEN_REPO_ID = "3"
REPO_BROKEN_TAG_RESPONSE = {"id": BROKEN_REPO_ID, "name": KOJI_BROKEN_REPO}
GET_REPO_RESPONSE = {"id": "2"}
ROOT = "http://example.com"


# ClientSession is xmlrpc instance, we need to mock it explicitly
class MockedClientSession(object):
    def __init__(self, hub, opts=None):
        self.ca_path = None
        self.cert_path = None
        self.serverca_path = None

    def getBuildTarget(self, target):
        if target == KOJI_TARGET_BROKEN_TAG:
            return BROKEN_TAG_RESPONSE
        if target == KOJI_TARGET_BROKEN_REPO:
            return BROKEN_REPO_RESPONSE
        return GET_TARGET_RESPONSE

    def getTag(self, tag):
        if tag == KOJI_BROKEN_TAG:
            return None
        if tag == KOJI_BROKEN_REPO:
            return REPO_BROKEN_TAG_RESPONSE
        return GET_TAG_RESPONSE

    def getRepo(self, repo):
        if repo == BROKEN_REPO_ID:
            return None
        return GET_REPO_RESPONSE

    def ssl_login(self, cert=None, ca=None, serverca=None, proxyuser=None):
        self.ca_path = ca
        self.cert_path = cert
        self.serverca_path = serverca
        return True

    def krb_login(self, *args, **kwargs):
        return True


class MockedPathInfo(object):
    def __init__(self, topdir=None):
        self.topdir = topdir

    def repo(self, repo_id, name):
        return "{0}/repos/{1}/{2}".format(self.topdir, name, repo_id)


def prepare():
    if MOCK:
        mock_docker()
    tasker = DockerTasker()
    workflow = DockerBuildWorkflow("test-image", source=SOURCE)
    workflow.source = StubSource()
    workflow.builder = StubInsideBuilder().for_workflow(workflow)

    session = MockedClientSession(hub='', opts=None)
    workflow.koji_session = session
    flexmock(koji,
             ClientSession=session,
             PathInfo=MockedPathInfo)

    return tasker, workflow


class TestKoji(object):
    @pytest.mark.parametrize('parent_images', [True, False])
    @pytest.mark.parametrize('base_from_scratch', [True, False])
    @pytest.mark.parametrize(('target', 'expect_success'), [
        (KOJI_TARGET, True),
        (KOJI_TARGET_BROKEN_TAG, False),
        (KOJI_TARGET_BROKEN_REPO, False)])
    @pytest.mark.parametrize(('root',
                              'koji_ssl_certs',
                              'expected_string',
                              'expected_file',
                              'proxy'), [
        # Plain http repo
        ('http://example.com',
         False,
         None,
         None,
         None),

        # Plain http repo with proxy
        ('http://example.com',
         False,
         None,
         None,
         'http://proxy.example.com'),

        # https with koji_ssl_certs
        # ('https://example.com',
        #  True,
        #  'sslcacert=',
        #  '/etc/yum.repos.d/example.com.cert'),

        # https with no cert available
        ('https://nosuchwebsiteforsure.com',
         False,
         'sslverify=0',
         None,
         None),

        # https with no cert available
        ('https://nosuchwebsiteforsure.com',
         False,
         'sslverify=0',
         None,
         'http://proxy.example.com'),

        # https with cert available
        # ('https://example.com',
        #  False,
        #  'sslcacert=/etc/yum.repos.d/example.com.cert',
        #  '/etc/yum.repos.d/example.com.cert'),

        # https with a cert for authentication
        ('https://nosuchwebsiteforsure.com',
         True,
         'sslverify=0',
         None,
         'http://proxy.example.com'),


    ])
    def test_koji_plugin(self, tmpdir, caplog, parent_images, base_from_scratch,
                         target, expect_success, root, koji_ssl_certs,
                         expected_string, expected_file, proxy, reactor_config_map):
        tasker, workflow = prepare()
        workflow.builder.base_from_scratch = base_from_scratch
        workflow.builder.parent_images = parent_images

        args = {'target': target}

        if koji_ssl_certs:
            tmpdir.join('cert').write('cert')
            tmpdir.join('serverca').write('serverca')

        if reactor_config_map:
            koji_map = {
                'hub_url': '',
                'root_url': root,
                'auth': {}}
            if koji_ssl_certs:
                koji_map['auth']['ssl_certs_dir'] = str(tmpdir)
            workflow.plugin_workspace[ReactorConfigPlugin.key] = {}
            workflow.plugin_workspace[ReactorConfigPlugin.key][WORKSPACE_CONF_KEY] =\
                ReactorConfig({'version': 1,
                               'koji': koji_map,
                               'yum_proxy': proxy})

        else:
            args.update({
                'hub': '',
                'root': root,
                'proxy': proxy,
            })

            if koji_ssl_certs:
                args['koji_ssl_certs_dir'] = str(tmpdir)

        runner = PreBuildPluginsRunner(tasker, workflow, [{
            'name': KojiPlugin.key,
            'args': args,
        }])

        runner.run()

        if base_from_scratch and not parent_images:
            log_msg = "from scratch single stage can't add repos from koji target"
            assert log_msg in caplog.text
            return
        if not expect_success:
            return

        if koji_ssl_certs:
            for file_path, expected in [(workflow.koji_session.cert_path, 'cert'),
                                        (workflow.koji_session.serverca_path, 'serverca')]:

                assert os.path.isfile(file_path)
                with open(file_path, 'r') as fd:
                    assert fd.read() == expected

        repofile = '/etc/yum.repos.d/target-?????.repo'
        assert len(workflow.files) == 1
        assert fnmatch(next(iter(workflow.files.keys())), repofile)
        content = next(iter(workflow.files.values()))
        assert content.startswith("[atomic-reactor-koji-plugin-target]\n")
        assert "gpgcheck=0\n" in content
        assert "enabled=1\n" in content
        assert "name=atomic-reactor-koji-plugin-target\n" in content
        assert "baseurl=%s/repos/tag/2/$basearch\n" % root in content

        if proxy:
            assert "proxy=%s" % proxy in content

        if expected_string:
            assert expected_string in content

        if expected_file:
            assert expected_file in workflow.files

    @pytest.mark.parametrize('target, yum_repos, include_repo', [
        ('target', ['repo'], False),
        (None, ['repo'], False),
        (None, ['repo'], True),
        (None, [], False),
        (None, [], True),
    ])
    def test_skip_plugin(self, caplog, target, yum_repos, include_repo):
        tasker, workflow = prepare()
        args = {'target': target}

        koji_map = {
            'hub_url': '',
            'root_url': 'http://example.com',
            'auth': {}}
        workflow.plugin_workspace[ReactorConfigPlugin.key] = {}
        workflow.plugin_workspace[ReactorConfigPlugin.key][WORKSPACE_CONF_KEY] =\
            ReactorConfig({'version': 1,
                           'koji': koji_map})

        workflow.user_params['include_koji_repo'] = include_repo
        workflow.user_params['yum_repourls'] = yum_repos

        runner = PreBuildPluginsRunner(tasker, workflow, [{
            'name': KojiPlugin.key,
            'args': args,
        }])

        runner.run()

        if (not include_repo and yum_repos):
            log_msg = 'there is a yum repo user parameter, skipping plugin'
        else:
            log_msg = 'no target provided, skipping plugin'

        assert log_msg in caplog.text
