from __future__ import unicode_literals

from psycopg2 import IntegrityError
from gratipay.testing import Harness, D
from gratipay.models.exchange_route import ExchangeRoute


class TestRecordAnExchange(Harness):

    # fixture
    # =======

    def make_participants(self):
        self.make_participant('alice', claimed_time='now', is_admin=True)
        self.bob = self.make_participant('bob', claimed_time='now')

    def record_an_exchange(self, data, make_participants=True):
        if make_participants:
            self.make_participants()

        data.setdefault('status', 'succeeded')
        data.setdefault('note', 'noted')

        if 'route_id' not in data:
            try:
                data['route_id'] = ExchangeRoute.insert(self.bob, 'paypal', 'bob@gmail.com').id
            except IntegrityError:
                data['route_id'] = ExchangeRoute.from_network(self.bob, 'paypal').id

        if data['status'] is None:
            del(data['status'])
        if data['route_id'] is None:
            del(data['route_id'])

        if 'ref' not in data:
            data['ref'] = 'N/A'

        return self.client.PxST('/~bob/history/record-an-exchange', data, auth_as='alice')


    # tests
    # =====

    def test_success_is_302(self):
        response = self.record_an_exchange({'amount': '10', 'fee': '0'})
        assert response.code == 302
        assert response.headers['location'] == '/bob/history/'

    def test_non_admin_is_403(self):
        self.make_participant('alice', claimed_time='now')
        self.bob = self.make_participant('bob', claimed_time='now')
        actual = self.record_an_exchange({'amount': '10', 'fee': '0'}, False).code
        assert actual == 403

    def test_bad_amount_is_400(self):
        response = self.record_an_exchange({'amount': 'cheese', 'fee': '0'})
        assert response.code == 400
        assert response.body == "Invalid amount/fee"

    def test_bad_fee_is_400(self):
        response = self.record_an_exchange({'amount': '10', 'fee': 'cheese'})
        assert response.code == 400
        assert response.body == "Invalid amount/fee"

    def test_no_note_is_400(self):
        response = self.record_an_exchange({'amount': '10', 'fee': '0', 'note': ''})
        assert response.code == 400
        assert response.body == "Invalid note"

    def test_whitespace_note_is_400(self):
        response = self.record_an_exchange({'amount': '10', 'fee': '0', 'note': '   '})
        assert response.code == 400
        assert response.body == "Invalid note"

    def test_no_route_id_is_400(self):
        response = self.record_an_exchange({'amount': '10', 'fee': '0', 'route_id': None})
        assert response.code == 400
        assert response.body == "Invalid route_id"

    def test_bad_route_id_is_400(self):
        response = self.record_an_exchange({'amount': '10', 'fee': '0', 'route_id': 'foo'})
        assert response.code == 400
        assert response.body == "Invalid route_id"

    def test_non_existent_route_id_is_400(self):
        response = self.record_an_exchange({'amount': '10', 'fee': '0', 'route_id': '123456'})
        assert response.code == 400
        assert response.body == "Route doesn't exist"

    def test_route_should_belong_to_user_else_400(self):
        alice = self.make_participant('alice', claimed_time='now', is_admin=True)
        self.make_participant('bob', claimed_time='now')
        route = ExchangeRoute.insert(alice, 'paypal', 'alice@gmail.com')

        response = self.record_an_exchange({'amount': '10', 'fee': '0', 'route_id': route.id}, False)
        assert response.code == 400
        assert response.body == "Route doesn't exist"

    def test_no_ref_is_400(self):
        response = self.record_an_exchange({'amount': '10', 'fee': '0', 'ref': ''})
        assert response.code == 400
        assert response.body == "Invalid Reference"

    def test_whitespace_ref_is_400(self):
        response = self.record_an_exchange({'amount': '10', 'fee': '0', 'ref': '   '})
        assert response.code == 400
        assert response.body == "Invalid Reference"

    def test_dropping_balance_below_zero_is_allowed_in_this_context(self):
        self.record_an_exchange({'amount': '-10', 'fee': '0'})
        actual = self.db.one("SELECT balance FROM participants WHERE username='bob'")
        assert actual == D('-10.00')

    def test_success_records_exchange(self):
        self.record_an_exchange({'amount': '10', 'fee': '0.50', 'ref':"605BSOC6G855L15OO"})
        expected = { "amount": D('10.00')
                   , "fee": D('0.50')
                   , "participant": "bob"
                   , "recorder": "alice"
                   , "note": "noted"
                   , "ref" : "605BSOC6G855L15OO"
                   , "route": ExchangeRoute.from_network(self.bob, 'paypal').id
                    }
        SQL = "SELECT amount, fee, participant, recorder, note, route, ref " \
              "FROM exchanges"
        actual = self.db.one(SQL, back_as=dict)
        assert actual == expected

    def test_success_updates_balance(self):
        self.record_an_exchange({'amount': '10', 'fee': '0'})
        expected = D('10.00')
        SQL = "SELECT balance FROM participants WHERE username='bob'"
        actual = self.db.one(SQL)
        assert actual == expected

    def test_withdrawals_work(self):
        self.make_participant('alice', claimed_time='now', is_admin=True)
        self.bob = self.make_participant('bob', claimed_time='now', balance=20)

        self.record_an_exchange({'amount': '-7', 'fee': '0'}, make_participants=False)

        expected = D('13.00')
        SQL = "SELECT balance FROM participants WHERE username='bob'"
        actual = self.db.one(SQL)
        assert actual == expected

    def test_withdrawals_take_fee_out_of_balance(self):
        self.make_participant('alice', claimed_time='now', is_admin=True)
        self.bob = self.make_participant('bob', claimed_time='now', balance=20)
        self.bob = self.record_an_exchange({'amount': '-7', 'fee': '1.13'}, False)
        SQL = "SELECT balance FROM participants WHERE username='bob'"
        assert self.db.one(SQL) == D('11.87')

    def test_can_set_status(self):
        self.make_participants()
        for status in ('pre', 'pending', 'failed', 'succeeded'):
            self.record_an_exchange({'amount': '10', 'fee': '0', 'status': status}, False)
            actual = self.db.one("SELECT status FROM exchanges ORDER BY timestamp desc LIMIT 1")
            assert actual == status

    def test_cant_record_new_exchanges_with_None_status(self):
        r = self.record_an_exchange({'amount': '10', 'fee': '0', 'status': None})
        assert r.code == 400
        assert self.db.one("SELECT count(*) FROM exchanges") == 0

    def test_succeeded_affects_balance(self):
        self.make_participants()
        balance = 0
        for amount in ('10', '-10'):
            self.record_an_exchange({'amount': amount, 'fee': '0'}, False)
            balance += int(amount)
            assert self.db.one("SELECT balance FROM participants WHERE username='bob'") == balance

    def test_failed_doesnt_affect_balance(self):
        self.make_participants()
        for amount in ('10', '-10'):
            self.record_an_exchange({
                'amount': amount,
                'fee': '0',
                'status': 'failed'
            }, False)
            assert self.db.one("SELECT balance FROM participants WHERE username='bob'") == 0

    def test_other_statuses_dont_affect_balance_for_payins(self):
        self.make_participants()
        for status in ('pre', 'pending'):
            self.record_an_exchange({
                'amount': '10',
                'fee': '0',
                'status': status
            }, False)
            assert self.db.one("SELECT balance FROM participants WHERE username='bob'") == 0

    def test_other_statuses_affect_balance_for_payouts(self):
        self.make_participants()
        balance = 0
        for status in ('pre', 'pending'):
            self.record_an_exchange({
                'amount': '-10',
                'fee': '0',
                'status': status
            }, False)
            balance -= 10
            assert self.db.one("SELECT balance FROM participants WHERE username='bob'") == balance
