import json
import os

import pytest

from opentrons.server import endpoints
from opentrons.config import pipette_config
from opentrons import config, types


@pytest.fixture
def restore_restart_required():
    yield
    endpoints.settings._SETTINGS_RESTART_REQUIRED = False


@pytest.fixture
async def attached_pipettes(async_client, request):
    """ Fixture the robot to have attached pipettes

    Mark the node with
    'attach_left_model': model_name for left (default: p300_single_v1)
    'attach_right_model': model_name for right (default: p50_multi_v1)
    'attach_left_id': id for left (default: 'abc123')
    'attach_right_id': id for right (default: 'acbcd123')

    Returns the model by mount style dict of
    {'left': {'name': str, 'model': str, 'id': str}, 'right'...}
    """
    def marker_with_default(marker: str, default: str) -> str:
        return request.node.get_closest_marker(marker) or default
    left_mod = marker_with_default('attach_left_model', 'p300_multi_v1')
    left_name = left_mod.split('_v')[0]
    right_mod = marker_with_default('attach_right_model', 'p50_multi_v1')
    right_name = right_mod.split('_v')[0]
    left_id = marker_with_default('attach_left_id', 'abc123')
    right_id = marker_with_default('attach_right_id', 'abcd123')
    hw = async_client.app['com.opentrons.hardware']
    if async_client.app['api_version'] == 1:
        hw.model_by_mount = {
            'left': {'model': left_mod, 'id': left_id, 'name': left_name},
            'right': {'model': right_mod, 'id': right_id, 'name': right_name}
        }
        hw.get_attached_pipettes()
        return hw.model_by_mount
    else:
        hw._backend._attached_instruments = {
            types.Mount.RIGHT: {
                'model': right_mod, 'id': right_id, 'name': right_name
            },
            types.Mount.LEFT: {
                'model': left_mod, 'id': left_id, 'name': left_name
            }
        }
        await hw.cache_instruments()
        return {k.name.lower(): v
                for k, v in hw._backend._attached_instruments.items()}


def validate_response_body(body):
    settings_list = body.get('settings')
    assert type(settings_list) == list
    for obj in settings_list:
        assert 'id' in obj, '"id" field not found in settings object'
        assert 'title' in obj, '"title" not found for {}'.format(obj['id'])
        assert 'description' in obj, '"description" not found for {}'.format(
            obj['id'])
        assert 'value' in obj, '"value" not found for {}'.format(obj['id'])
        assert 'restart_required' in obj
    assert 'links' in body
    assert isinstance(body['links'], dict)


async def test_get(async_client):
    resp = await async_client.get('/settings')
    body = await resp.json()
    assert resp.status == 200
    validate_response_body(body)
    if async_client.app['api_version'] == 2:
        assert 'enableApi1BackCompat' in [s['id'] for s in body['settings']]
    else:
        assert 'enableApi1BackCompat' not in [
            s['id'] for s in body['settings']]


async def test_set(virtual_smoothie_env, loop, async_client):
    test_id = 'disableHomeOnBoot'

    resp = await async_client.post(
        '/settings', json={"id": test_id, "value": True})
    body = await resp.json()
    assert resp.status == 200
    validate_response_body(body)
    test_setting = list(
        filter(lambda x: x.get('id') == test_id, body.get('settings')))[0]
    assert test_setting.get('value')


async def test_restart_required_persistence(
        virtual_smoothie_env, loop, async_client,
        restore_restart_required):
    current = await async_client.get('/settings')
    current_body = await current.json()
    assert not current_body['links']

    # Changing the state sets restart required
    cur_apiv2_state = [s['value'] for s in current_body['settings']
                       if s['id'] == 'useProtocolApi2'][0]
    new_apiv2_state = not bool(cur_apiv2_state)
    resp = await async_client.post('/settings',
                                   json={'id': 'useProtocolApi2',
                                         'value': new_apiv2_state})
    assert resp.status == 200
    body = await resp.json()
    assert body['links']['restart'] == '/server/restart'

    # Setting the state to the same value again doesn't clear it
    new = await async_client.get('/settings')
    new_body = await new.json()
    assert new_body['links']['restart'] == '/server/restart'

    resp = await async_client.post('/settings',
                                   json={'id': 'useProtocolApi2',
                                         'value': new_apiv2_state})
    assert resp.status == 200
    body = await resp.json()
    assert body['links']['restart'] == '/server/restart'

    # Neither does changing the state back
    new = await async_client.get('/settings')
    new_body = await new.json()
    assert new_body['links']['restart'] == '/server/restart'

    resp = await async_client.post('/settings',
                                   json={'id': 'useProtocolApi2',
                                         'value': cur_apiv2_state})
    assert resp.status == 200
    body = await resp.json()
    assert body['links']['restart'] == '/server/restart'


async def test_available_resets(async_client):
    resp = await async_client.get('/settings/reset/options')
    body = await resp.json()
    options_list = body.get('options')
    assert resp.status == 200
    if async_client.app['api_version'] == 1:
        options = [{'id': 'tipProbe',
                    'name': 'Tip Length',
                    'description': 'Clear tip probe data'},
                   {'id': 'labwareCalibration',
                    'name': 'Labware Calibration',
                    'description': 'Clear labware calibration'},
                   {'id': 'bootScripts',
                    'name': 'Boot Scripts',
                    'description': 'Clear custom boot scripts'}]
    else:
        options = [
            {'id': 'customLabware',
             'name': 'Custom Labware',
             'description': 'Clear custom labware definitions'},
            {'id': 'tipProbe',
             'name': 'Instrument Offset',
             'description':
             'Clear instrument offset calibration and tip probe data'},
            {'id': 'labwareCalibration',
             'name': 'Labware Calibration',
             'description': 'Clear labware calibration'},
            {'id': 'bootScripts',
             'name': 'Boot Scripts',
             'description': 'Clear custom boot scripts'}]
    assert sorted(options_list, key=lambda el: el['id'])\
        == sorted(options, key=lambda el: el['id'])


@pytest.mark.api1_only
async def execute_reset_tests_v1(async_client):
    # Make sure we actually delete the database
    cli = async_client
    resp = await cli.post('/settings/reset', json={'labwareCalibration': True})
    body = await resp.json()
    assert not os.path.exists(config.CONFIG['labware_database_file'])
    assert resp.status == 200
    assert body == {}

    # Make sure this one is idempotent
    resp = await cli.post('/settings/reset', json={'labwareCalibration': True})
    body = await resp.json()
    assert resp.status == 200
    assert body == {}

    # Check that we properly delete only the tip length key
    resp = await cli.post('/settings/reset', json={'tipProbe': True})
    body = await resp.json()
    assert resp.status == 200
    assert body == {}

    robot_settings = config.CONFIG['robot_settings_file']
    with open(robot_settings, 'r') as f:
        data = json.load(f)
    assert data['tip_length'] == {}

    # Check the inpost validation
    resp = await cli.post('/settings/reset', json={'aksgjajhadjasl': False})
    body = await resp.json()
    assert resp.status // 100 == 4
    assert 'message' in body
    assert 'aksgjajhadjasl' in body['message']


async def execute_reset_tests_v2(async_client):
    cli = async_client
    # Make sure we actually delete the database
    resp = await cli.post('/settings/reset', json={'labwareCalibration': True})
    body = await resp.json()
    assert not os.listdir(config.CONFIG['labware_calibration_offsets_dir_v2'])
    assert resp.status == 200
    assert body == {}

    # Make sure this one is idempotent
    resp = await cli.post('/settings/reset', json={'labwareCalibration': True})
    body = await resp.json()
    assert resp.status == 200
    assert body == {}

    # Check that we properly delete only the tip length key
    resp = await cli.post('/settings/reset', json={'tipProbe': True})
    body = await resp.json()
    assert resp.status == 200
    assert body == {}

    robot_settings = config.CONFIG['robot_settings_file']
    with open(robot_settings, 'r') as f:
        data = json.load(f)
    assert data['instrument_offset']\
        == config.robot_configs.build_fallback_instrument_offset({})

    # Check the inpost validation
    resp = await cli.post('/settings/reset', json={'aksgjajhadjasl': False})
    body = await resp.json()
    assert resp.status // 100 == 4
    assert 'message' in body
    assert 'aksgjajhadjasl' in body['message']


async def test_receive_pipette_settings(async_client, attached_pipettes):

    test_id = attached_pipettes['left']['id']
    resp = await async_client.get('/settings/pipettes')
    body = await resp.json()
    assert test_id in body
    assert body[test_id]['fields'] == pipette_config.list_mutable_configs(
        pipette_id=test_id)


async def test_receive_pipette_settings_one_pipette(
        async_client, attached_pipettes):
    # This will check that sending a known pipette id works,
    # and sending an unknown one does not
    test_id = attached_pipettes['left']['id']
    resp = await async_client.get('/settings/pipettes/{}'.format(test_id))
    body = await resp.json()
    assert body['fields'] == pipette_config.list_mutable_configs(
        pipette_id=test_id)

    # Non-existent pipette id and get 404
    resp = await async_client.get(
        '/settings/pipettes/{}'.format('wannabepipette'))
    assert resp.status == 404


async def test_modify_pipette_settings(async_client, attached_pipettes):
    # This test will check that setting modified pipette configs
    # works as expected
    changes = {
        'fields': {
            'pickUpCurrent': {'value': 1}
        }
    }

    no_changes = {
        'fields': {
            'pickUpCurrent': {'value': 1}
            }
        }

    test_id = attached_pipettes['left']['id']
    # Check data has not been changed yet
    resp = await async_client.get('/settings/pipettes/{}'.format(test_id))
    body = await resp.json()
    assert body['fields']['pickUpCurrent'] == \
        pipette_config.list_mutable_configs(
            pipette_id=test_id)['pickUpCurrent']

    # Check that data is changed and matches the changes specified
    resp = await async_client.patch(
        '/settings/pipettes/{}'.format(test_id),
        json=changes)
    patch_body = await resp.json()
    assert resp.status == 200
    check = await async_client.get('/settings/pipettes/{}'.format(test_id))
    body = await check.json()
    assert body['fields'] == patch_body['fields']

    # Check that None reverts a setting to default
    changes2 = {
        'fields': {
            'pickUpCurrent': None
            }
        }
    resp = await async_client.patch(
        '/settings/pipettes/{}'.format(test_id),
        json=changes2)
    assert resp.status == 200
    check = await async_client.get('/settings/pipettes/{}'.format(test_id))
    body = await check.json()
    assert body['fields']['pickUpCurrent']['value'] == \
        pipette_config.list_mutable_configs(
            pipette_id=test_id)['pickUpCurrent']['default']

    # check no fields returns no changes
    resp = await async_client.patch(
        '/settings/pipettes/{}'.format(test_id),
        json=no_changes)
    body = await resp.json()
    assert body['fields'] == pipette_config.list_mutable_configs(test_id)
    assert resp.status == 200


async def test_incorrect_modify_pipette_settings(
        async_client, attached_pipettes):
    out_of_range = {
            'fields': {
                'pickUpCurrent': {'value': 1000}
                }
            }
    # check over max fails
    resp = await async_client.patch(
        '/settings/pipettes/{}'.format(attached_pipettes['left']['id']),
        json=out_of_range)
    assert resp.status == 412


async def test_set_log_level(async_client):
    # Check input sanitization
    hardware = async_client.app['com.opentrons.hardware']
    resp = await async_client.post('/settings/log_level/local', json={})
    assert resp.status == 400
    body = await resp.json()
    assert 'message' in body
    resp = await async_client.post('/settings/log_level/local',
                                   json={'log_level': 'oafajhshda'})
    assert resp.status == 400
    body = await resp.json()
    assert 'message'in body
    if async_client.app['api_version'] == 1:
        conf = hardware.config
    else:
        conf = await hardware.config
    assert conf.log_level != 'ERROR'
    resp = await async_client.post('/settings/log_level/local',
                                   json={'log_level': 'error'})
    assert resp.status == 200
    body = await resp.json()
    assert 'message' in body
    if async_client.app['api_version'] == 1:
        conf = hardware.config
    else:
        conf = await hardware.config
    assert conf.log_level == 'ERROR'
