#  Copyright CERFACS (http://cerfacs.fr/)
#  Apache License, Version 2.0 (http://www.apache.org/licenses/LICENSE-2.0)
#
#  Author: Natalia Tatarinova
#  Additions (2015, 2016): Christian Page

import numpy
import sys
import pdb
from datetime import timedelta
from netCDF4 import Dataset

if sys.version_info[0] < 3:
    import util_dt
else:
    from icclim.util import util_dt


def read_netCDF(file):

    try:
        inc = Dataset(file, 'r')
    except RuntimeError:
        raise MissingIcclimInputError("Failed to access dataset: " + file)
    return inc

def create_output_netcdf(netcdf_version, out_file):

    netcdfv = ['NETCDF4', 'NETCDF4_CLASSIC', 'NETCDF3_CLASSIC', 'NETCDF3_64BIT']
    if netcdf_version not in netcdfv:
        netcdf_version = icclim_output_file_defaults('netcdf_version')
    onc = Dataset(out_file, 'w' ,format=netcdf_version)
    return onc


def set_date_event(onc, user_indice, indice_dim, fill_val, ncVar_time):

    if user_indice['calc_operation'] in ['min', 'max']:            
        date_event = onc.createVariable('date_event', 'f', indice_dim, fill_value = fill_val)
        # we set the same 'calendar' and 'units' attributes as those of netCDF var 'time'
        date_event.__setattr__('calendar', ncVar_time.calendar)
        date_event.__setattr__('units', ncVar_time.units)
        
    elif user_indice['calc_operation'] in ['nb_events', 'max_nb_consecutive_events', 'run_mean', 'run_sum']:
        date_event_start = onc.createVariable('date_event_start', 'f', indice_dim, fill_value = fill_val)
        # we set the same 'calendar' and 'units' attributes as those of netCDF var 'time'
        date_event_start.__setattr__('calendar', ncVar_time.calendar)
        date_event_start.__setattr__('units', ncVar_time.units)
        
        date_event_end = onc.createVariable('date_event_end', 'f', indice_dim, fill_value = fill_val)
        # we set the same 'calendar' and 'units' attributes as those of netCDF var 'time'
        date_event_end.__setattr__('calendar', ncVar_time.calendar)
        date_event_end.__setattr__('units', ncVar_time.units)


def set_threshold(onc, threshold, indice_dim):

    # Create an extra dimension for the index:
    indice_dim.insert(1,'threshold')
    onc.createDimension('threshold',nb_user_thresholds)
    thresholdvar = onc.createVariable('threshold','f8',('threshold'))
    thresholdvar[:] = user_thresholds
    thresholdvar.setncattr("units","threshold")
    thresholdvar.setncattr("standard_name","threshold")


def check_att(nc, att):
    
    '''    
    Checks if a global attribut exists in a dataset.
    
    :param nc: NetCDF dataset
    :type nc: netCDF4.Dataset 
    :param att: attribut name
    :type att: str
    
    :rtype: int (1 if attribut exists, 0 else)
    
    '''
    
    try:
        nc.__getattribute__(att)
        a = 1 # attribut exists
    except AttributeError:
        a = 0 # attribut doesn't exist
    return a


def copy_att(nc1, nc2, att):
    
    '''
    Copies a global attribute from one dataset (nc1) to another (nc2).
    
    :param nc1: from
    :type nc1: netCDF4.Dataset
    :param nc2: to
    :type nc2: netCDF4.Dataset

    '''
    
    nc2.__setattr__(att, '')
    if (check_att(nc1,att)==1):
        nc2.__setattr__(att, nc1.__getattribute__(att))
        
        
def copy_var_attrs(source, destination):
    '''
    Copies all attributes from one variable (source) to another (destination).
    Except scale_factor add_offset and _FillValue...
    
    :param source: from
    :type source: netCDF4.Variable
    :param destination: to
    :type destination: netCDF4.Variable

    '''
    sourceAttrs = source.ncattrs()
    for attr in sourceAttrs:
        if attr == '_FillValue' or attr == 'scale_factor' or attr == 'add_offset':
            pass
        else:
            destination.setncattr(attr, source.getncattr(attr))
            
            
def copy_var(variableName,sourceDataset, destinationDataset):
    '''
    Copies a variable from one dataset(sourceDataset) to another(destinationDataset) by name(variableName).
    
    :param variableName: name of variable to write
    :type variableName: str
    :param sourceDataset: from
    :type sourceDataset: netCDF4.Dataset
    :param destinationDataset: to
    :type destinationDataset: netCDF4.Dataset

    '''
    # Get the variable to copy
    try:
      sourceVar = sourceDataset.variables[variableName];
    
      # Copy the dims of the variable
      for dimname in sourceVar.dimensions:
          if dimname not in destinationDataset.dimensions:
              dim = rootgrp.dimensions.get(dimname)
              destinationDataset.createDimension(dimname,len(dim))
              
      # Create the variable
      destinationVariable = destinationDataset.createVariable(variableName,sourceVar.dtype, sourceVar.dimensions)
    
      #Copy its attributes
      copy_var_attrs(sourceVar,destinationVariable);
      destinationVariable[:] = sourceVar[:]
    except KeyError:
      sourceVar = None
    
def get_att_value(nc, var, att):
    '''
    Returns an attribut value of a variable in dataset.
    
    :param nc: NetCDF dataset
    :type nc: netCDF4.Dataset
    :param var: variable name
    :type var: str
    :param att: attribut name
    :type var: str
    
    '''
    v = nc.variables[var]
    att_val = v.getncattr(att)
    return att_val


def set_time_values(nc, time_steps_arr_dt, calend, units):
    '''
    param time_steps_arr_dt: numpy array of datetime objects
    '''
    time_steps_num = numpy.array([util_dt.date2num(i, calend, units) for i in time_steps_arr_dt])
    nc.variables['time'][:] = time_steps_num[:]

def set_timebnds_values(nc, time_bnds_dt, calend, units):
    time_bnds_num = numpy.array([util_dt.date2num(i, calend, units) for i in time_bnds_dt])
    nc.variables['time_bnds'][:,:] = time_bnds_num[:,:]
    
    
def copy_var_dim(inc, onc, var, lev_dim_pos=1): 
    '''
    Copies the spatial coordinate variables (e.g.: lat, lon) of a variable (var) from one NetCDF file (ifile) to another (out_file)
    and returns list of coordinates variables.
    
    :param inc: input dataset
    :type inc: netCDF4.Dataset
    :param onc: output dataset
    :type onc: netCDF4.Dataset
    :param var: variable name to process
    :type var: str
    :param project: project name ("CMIP5" or "CORDEX")
    :type project: str
    :param lev_dim_pos: Position of Level dimension, either 0 or 1. 0 is leftmost dimension, 1 is second to the leftmost. Default 1.
    :type lev_dim_pos: int
    
    :rtype: tuple of str (coordinate variables: 'time', 'lat', 'lon')
    '''


    v = inc.variables[var]

    v_dim = v.dimensions
    
    if v.ndim == 3: # (e.g.: u'time', u'lat', u'lon')
        time_var = v_dim[0]
        lat_var = v_dim[1]
        lon_var = v_dim[2]
        lat_shape = v.shape[1]
        lon_shape = v.shape[2]
        
    elif v.ndim == 4: # (e.g.: u'time', u'plev', u'lat', u'lon')
        if lev_dim_pos == 0:
            time_var = v_dim[1]
        else:
            time_var = v_dim[0]
        lat_var = v_dim[-2]
        lon_var = v_dim[-1]
        lat_shape = v.shape[-2]
        lon_shape = v.shape[-1]
        
    glob_att = ['title', 'institution', 'source', 'references', 'comment', 'history']
    
    # Note: it needs to do it BEFORE the creation of var/dim
    for att in glob_att:
        copy_att(inc,onc,att)
       
    
    
        
    inc_dim0 = inc.variables[str(time_var)] # time
    inc_dim1 = inc.variables[str(lat_var)] # lat
    inc_dim2 = inc.variables[str(lon_var)] # lon

    try:
        inc_dim1_bounds = inc.variables[str(inc_dim1.__getattribute__('bounds'))]
        inc_dim2_bounds = inc.variables[str(inc_dim2.__getattribute__('bounds'))]
        boundsvar = 1
    except:
        boundsvar = 0
        pass

    
    ### create dimensions
    onc.createDimension(str(time_var), 0) # time       
    onc.createDimension(str(lat_var), lat_shape) # lat
    onc.createDimension(str(lon_var), lon_shape) # lon        
    onc.createDimension('bnds', 2) # bnds
    if boundsvar == 1:
      onc.createDimension('nv', 2) # nv
    
    ### create variables 
    onc_dim0 = onc.createVariable( str(time_var), inc_dim0.dtype, (str(time_var)) ) # time
    onc_dim1 = onc.createVariable( str(lat_var), inc_dim1.dtype, (str(lat_var)) ) # lat
    onc_dim2 = onc.createVariable( str(lon_var), inc_dim2.dtype, (str(lon_var)) ) # lon
    onc_time_bnds = onc.createVariable( 'time_bnds', inc_dim0.dtype, (str(time_var), 'bnds') ) # time_bnds

    if boundsvar == 1:
      try:
          onc_dim1_bounds = onc.createVariable('lat_bnds', inc_dim1_bounds.dtype, (str(lat_var), 'nv'))
          onc_dim2_bounds = onc.createVariable('lon_bnds', inc_dim2_bounds.dtype, (str(lon_var), 'nv'))
      except:
          pass


    ### copy attributes // work for python 2 & 3
    # Change has been done cause Orderdict type doesn't support index in python 3         
    # time
    for attr in inc_dim0.__dict__.items():
        onc_dim0.__setattr__(attr[0], attr[1])
    # lat
    for attr in inc_dim1.__dict__.items(): # set attributs of current variable
        onc_dim1.__setattr__(attr[0]  , attr[1])  
    # lon
    for attr in inc_dim2.__dict__.items(): # set attributs of current variable       
        onc_dim2.__setattr__(attr[0]  , attr[1])

    try:
        # lat_bnds
        for attr in inc_dim1_bounds.__dict__.items(): # set attributs of current variable       
            onc_dim1_bounds.__setattr__(  attr[0]  , attr[1])
    
        # lon_bnds
        for attr in inc_dim2_bounds.__dict__.items(): # set attributs of current variable       
            onc_dim2_bounds.__setattr__(  attr[0]  , attr[1])
    except:
        pass        

    
    # for time_bnds, we copy only 2 attributes of time: 'units' and 'calendar' ( => var 'time' must have these 2 attributes)
    onc_time_bnds.__setattr__( 'units', inc_dim0.__getattribute__('units') )
    try:
        onc_time_bnds.__setattr__( 'calendar', inc_dim0.__getattribute__('calendar') )
    except:
        onc_time_bnds.__setattr__( 'calendar', 'gregorian' )
    
    
    ### copy values
    onc_dim1[:] = inc_dim1[:]
    onc_dim2[:] = inc_dim2[:]
    try:
        onc_dim1_bounds[:,:] = inc_dim1_bounds[:,:]
        onc_dim2_bounds[:,:] = inc_dim2_bounds[:,:]
    except:
        pass
    #######################
    
    # Copy coordinate variables
    if check_att(v, 'coordinates')==1:
      #print "There are coordinate vars"
      coordinateAttr = v.getncattr("coordinates").split()
      for coordinate in coordinateAttr:
        #print "Copying coordidinate"+coordinate
          copy_var(coordinate,inc,onc)
    # maartenplieger comments: Is this necessary? Should copy all grid_mapping of the var to write and it will be fine.
    if check_att(v, 'grid_mapping')==1:
          c = str(v.__getattribute__('grid_mapping'))
          inc_c = inc.variables[c]
          
          onc_c = onc.createVariable( c, inc.variables[c].dtype )

          for attr in inc_c.__dict__.items(): # set attributs of current variable       
              onc_c.__setattr__(  attr[0]  , attr[1])
    
    return (str(time_var), str(lat_var), str(lon_var)) # tuple ('time', 'lat', 'lon')


def get_values_arr_and_dt_arr(ncVar_temporal, ncVar_values, fill_val=None, time_range=None, N_lev=None, lev_dim_pos=1, ignore_Feb29th=False, i1_row_current_tile=None, i2_row_current_tile=None, i1_col_current_tile=None, i2_col_current_tile=None, add_offset=0.0, scale_factor=1.0):
    
    try:
        calend = ncVar_temporal.calendar
    except:
        calend = 'gregorian'
    units=ncVar_temporal.units
    
    time_arr = ncVar_temporal[:]
    
    dt_arr = numpy.array([util_dt.num2date(dt, calend=calend, units=units) for dt in time_arr])

    # REMOVED, because netcdftime.datetime objects have no method total_seconds()
#     deltat = (dt_arr[1]-dt_arr[0]).total_seconds()
#     if deltat != 86400.0:
#         print "WARNING: Time interval of the input file is not daily!! Delta time is: "+str(deltat)
    
#     print  "+++", time_range   
    if N_lev == None:
        assert(ncVar_values.ndim == 3)  
                       
        indices_subset = util_dt.get_indices_subset(dt_arr, time_range)
        dt_arr = dt_arr[indices_subset]
        values_arr = (ncVar_values[indices_subset,i1_row_current_tile:i2_row_current_tile, i1_col_current_tile:i2_col_current_tile] * scale_factor) + add_offset
            
    else:
        assert(ncVar_values.ndim == 4)
            
        indices_subset = util_dt.get_indices_subset(dt_arr, time_range)
        dt_arr = dt_arr[indices_subset]
        if lev_dim_pos == 0:
            values_arr = (ncVar_values[N_lev,indices_subset,i1_row_current_tile:i2_row_current_tile, i1_col_current_tile:i2_col_current_tile] * scale_factor) + add_offset
        else:
            values_arr = (ncVar_values[indices_subset,N_lev,i1_row_current_tile:i2_row_current_tile, i1_col_current_tile:i2_col_current_tile] * scale_factor) + add_offset
        
    if fill_val != None:
        numpy.ma.set_fill_value(values_arr, fill_val)
        
    assert(dt_arr.ndim == 1)    
    assert(values_arr.ndim == 3)
    
    if ignore_Feb29th == True and not calend == '360_day':
        mask_Feb29th = numpy.array([ (dt.month==2 and dt.day==29) for dt in dt_arr])
        indices_masked_Feb29th = numpy.where(mask_Feb29th==False)[0] # ...[0]: tuple to numpy.ndarray (http://stackoverflow.com/questions/16127444/why-is-my-array-length-1-when-building-it-with-numpy-where)
        dt_arr = dt_arr[indices_masked_Feb29th]
        values_arr = values_arr[indices_masked_Feb29th,:,:]

    #values_arr = values_arr.astype(numpy.float64)
    
    return (dt_arr, values_arr)

def list_var_dim(inc, var): 
    '''
    Get the spatial coordinate variables (e.g.: lat, lon) of a variable (var) from one NetCDF file (ifile) and returns list of coordinates variables.
    
    :param inc: input dataset
    :type inc: netCDF4.Dataset
    :param var: variable name to process
    :type var: str
    :param project: project name ("CMIP5" or "CORDEX")
    :type project: str
    
    :rtype: tuple of str (coordinate variables: 'time', 'lat', 'lon')
    '''


    v = inc.variables[var]

    v_dim = v.dimensions
    
    if v.ndim == 3: # (e.g.: u'time', u'lat', u'lon')
        time_var = v_dim[0]
        lat_var = v_dim[1]
        lon_var = v_dim[2]
        
    elif v.ndim == 4: # (e.g.: u'time', u'plev', u'lat', u'lon')
        time_var = v_dim[0]
        lat_var = v_dim[-2]
        lon_var = v_dim[-1]

                
    return (str(time_var), str(lat_var), str(lon_var)) # tuple ('time', 'lat', 'lon')

def check_unlimited(infile):
    
    '''    
    Checks for unlimited dimensions in a NetCDF file.
    
    :param infile: name of NetCDF file
    :type infile: str
    
    :rtype: str (name of unlimited dimension, defaulting to time in case there is none)
    
    '''
    
    dim_name = 'time'
    num_unlimited = 0
    dim_unlimited = False

    nc = Dataset(infile, 'r')

    for dim in nc.dimensions:
      if nc.dimensions[dim].isunlimited():
        dim_unlimited = True
        dim_name = dim
        num_unlimited = num_unlimited + 1
    if dim_unlimited == False:
      print('Warning: There is no unlimited dimension. File should be fixed if possible to set time as the unlimited dimension.')
      print('Warning: Using time as the aggregation dimension. Hope this is what you want to do...')
    if num_unlimited > 1:
        print('Warning: There are more than one unlimited dimension to aggregate. Using '+dim_name+' to aggregate.')

    nc.close()

    return dim_name

def save_percentile_netcdf(out_file, percentile_array):

    onc = Dataset(out_file[:-3]+"percentile_array.nc", 'w' ,format='NETCDF4_CLASSIC')

    time = onc.createDimension('time', percentile_array.shape[0])
    lat = onc.createDimension('lat', percentile_array.shape[1])
    lon = onc.createDimension('lon', percentile_array.shape[2])

    Perc = onc.createVariable('Perc', numpy.float32, ('time', 'lat','lon')) 

    Perc[:] = percentile_array

    onc.close()